""" Utility functions for generating test fixtures """

import time
from uuid import uuid4

from pennsieve import Pennsieve


def current_ts():
    """ Gets current timestamp """
    return int(round(time.time() * 1000))


def get_test_client(profile=None, api_token=None, api_secret=None, **overrides):
    """ Utility function to get a Pennsieve client object """
    ps = Pennsieve(
        profile=profile, api_token=api_token, api_secret=api_secret, **overrides
    )
    assert ps.context is not None
    orgs = ps.organizations()
    assert len(orgs) > 0

    # explicitly set context to Blackfyn org
    assert ps.context is not None

    return ps


def create_test_dataset(ps_client):
    """Utility function to generate a dataset for testing. It is up to the
    caller to ensure the dataset is cleaned up
    """
    ds = ps_client.create_dataset("test_dataset_{}".format(uuid4()))
    ds_id = ds.id
    # Removing this check to limit the number of API calls:
    # all_dataset_ids = [x.id for x in ps_client.datasets()]
    # assert ds_id in all_dataset_ids
    return ds



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from pennsieve.api.base import APIBase


class UserAPI(APIBase):
    base_uri = "/user"
    name = "user"

    def switch_organization(self, orgId):
        return self._put(self._uri("/organization/{orgId}/switch".format(orgId=orgId)))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cache_segment.proto
# isort:skip_file

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="cache_segment.proto",
    package="com.pennsieve",
    syntax="proto3",
    serialized_pb=_b(
        '\n\x13\x63\x61\x63he_segment.proto\x12\rcom.pennsieve">\n\x0c\x43\x61\x63heSegment\x12\x11\n\tchannelId\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x62\x06proto3'
    ),
)


_CACHESEGMENT = _descriptor.Descriptor(
    name="CacheSegment",
    full_name="com.pennsieve.CacheSegment",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="channelId",
            full_name="com.pennsieve.CacheSegment.channelId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="index",
            full_name="com.pennsieve.CacheSegment.index",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="com.pennsieve.CacheSegment.data",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=38,
    serialized_end=100,
)

DESCRIPTOR.message_types_by_name["CacheSegment"] = _CACHESEGMENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CacheSegment = _reflection.GeneratedProtocolMessageType(
    "CacheSegment",
    (_message.Message,),
    dict(
        DESCRIPTOR=_CACHESEGMENT,
        __module__="cache_segment_pb2"
        # @@protoc_insertion_point(class_scope:com.pennsieve.CacheSegment)
    ),
)
_sym_db.RegisterMessage(CacheSegment)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from builtins import dict, object
from future.utils import string_types

import io
import os
import sys
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed

from pennsieve import log
from pennsieve.api.agent import agent_upload, validate_agent_installation
from pennsieve.api.base import APIBase
from pennsieve.models import Collection, DataPackage, Dataset, TimeSeries

# GLOBAL
UPLOADS = {}

logger = log.get_logger("pennsieve.api.transfers")


def check_files(files):
    for f in files:
        if not os.path.exists(f):
            raise Exception("File {} not found.".format(f))


class IOAPI(APIBase):
    """
    Input/Output interface.
    """

    name = "io"

    def upload_files(
        self,
        destination,
        files,
        dataset=None,
        append=False,
        display_progress=False,
        recursive=False,
    ):
        if isinstance(destination, Dataset):
            # uploading into dataset
            destination_id = None
            dataset = destination
            dataset_id = self._get_id(dataset)
        elif destination is None and dataset is not None:
            # uploading into dataset
            destination_id = None
            dataset_id = self._get_id(dataset)
        elif isinstance(destination, Collection):
            # uploading into collection
            destination_id = self._get_id(destination)
            dataset_id = self._get_id(destination.dataset)
        elif append and isinstance(destination, TimeSeries):
            # uploading into timeseries package must be an append
            destination_id = self._get_id(destination)
            dataset_id = self._get_id(destination.dataset)
        elif isinstance(destination, string_types):
            # assume ID is for collection
            if dataset is None:
                raise Exception(
                    "Must also supply dataset when specifying destination by ID"
                )
            destination_id = destination
            dataset_id = self.get_id(dataset)
        else:
            raise Exception(
                "Cannot upload to destination of type {}".format(type(destination))
            )

        # check input files
        check_files(files)

        # sanity check dataset
        try:
            if isinstance(dataset, Dataset) and dataset.exists:
                ds = dataset
            else:
                ds = self.session.datasets.get(dataset_id)
        except:
            raise Exception("dataset does not exist")

        if destination_id is not None:
            try:
                if (
                    isinstance(destination, (Dataset, Collection, DataPackage))
                    and destination.exists
                ):
                    pass
                else:
                    destination = self.session.packages.get(destination_id)
            except:
                raise Exception("destination does not exist")

            # check type for appends
            if append:
                if not isinstance(destination, TimeSeries):
                    raise Exception("Append destination must be TimeSeries package.")
            else:
                if not isinstance(destination, Collection):
                    raise Exception("Upload destination must be Collection or Dataset.")

        # Push uploads through the Pennsieve agent
        validate_agent_installation(self.session.settings)

        return agent_upload(
            destination=destination,
            files=files,
            dataset=ds,
            append=append,
            recursive=recursive,
            display_progress=display_progress,
            settings=self.session.settings,
        )



from __future__ import absolute_import, division, print_function

import logging
import os

logging.basicConfig()


def get_logger(name="pennsieve-python"):
    """
    Returns a logger configured to be used throughout the
    pennsieve-python library

    Args:
      name (string): Name of the logger, default "pennsieve-python"

    Returns:
      Logger
    """

    logger = logging.getLogger(name)
    logger.setLevel(get_log_level())

    return logger


def get_log_level(default="INFO"):
    return os.environ.get("PENNSIEVE_LOG_LEVEL", default)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function
from future.utils import integer_types, string_types

import datetime

from pennsieve.extensions import numpy as np
from pennsieve.extensions import pandas as pd
from pennsieve.extensions import require_extension

# data type helpers


def value_as_type(value, dtype):
    try:
        if dtype == "string":
            return str(value)
        elif dtype == "integer":
            return int(value)
        elif dtype == "double":
            return float(value)
        elif dtype == "date":
            return infer_epoch_msecs(value)
        if dtype == "boolean":
            return value.lower() == "true"
    except:
        raise Exception("Unable to set value={} as type {}".format(value, dtype))


def get_data_type(v):
    """
    Infers type from value. Returns tuple of (type, value)
    """
    if isinstance(v, datetime.datetime):
        return ("date", msecs_since_epoch(v))
    elif isinstance(v, bool):
        return ("boolean", str(v).lower())
    elif isinstance(v, float):
        return ("double", v)
    elif isinstance(v, integer_types):
        return ("integer", v)
    else:
        # infer via casting
        if is_integer(v):
            return ("integer", int(v))
        elif is_decimal(v):
            return ("double", float(v))
        else:
            return ("string", str(v))


def is_integer(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def is_decimal(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


# time-series helpers


def infer_epoch_msecs(thing):
    if isinstance(thing, datetime.datetime):
        return msecs_since_epoch(thing)
    elif isinstance(thing, (integer_types, float)):
        # assume milliseconds
        return int(thing)
    elif isinstance(thing, string_types):
        # attempt to convert to msec integer
        return int(thing)
    else:
        raise Exception("Cannot parse date")


def infer_epoch(thing):
    if isinstance(thing, datetime.datetime):
        return usecs_since_epoch(thing)
    elif isinstance(thing, (integer_types, float)):
        # assume microseconds
        return int(thing)
    else:
        raise Exception("Cannot parse date")


def secs_since_epoch(the_time):
    the_time = the_time.replace(tzinfo=None)
    # seconds from epoch (float)
    return (the_time - datetime.datetime.utcfromtimestamp(0)).total_seconds()


def msecs_since_epoch(the_time):
    # milliseconds from epoch (integer)
    return int(secs_since_epoch(the_time) * 1000)


def usecs_since_epoch(the_time):
    # microseconds from epoch (integer)
    return int(secs_since_epoch(the_time) * 1e6)


def usecs_to_datetime(us):
    # convert usecs since epoch to proper datetime object
    return datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(
        microseconds=int(us)
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Timeseries helpers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@require_extension
def generate_data(size, func="walk", scale=5, periods=10):

    remainder = size % periods
    pattern_length = int(size / periods)
    if func == "walk":
        x = (np.random.rand(size) - 0.5).cumsum()
        # normalize to [-scale,scale]
        return x / np.max([x.min(), x.max()]) * scale
    elif func == "sin":
        return np.sin(np.linspace(0, np.pi * periods, size)) * scale
    elif func == "square":
        pattern = np.concatenate(
            [np.ones(pattern_length // 2) * -1, np.ones(pattern_length // 2)]
        )
        return (
            np.concatenate([np.repeat(pattern, periods), pattern[:remainder]]) * scale
        )
    elif func == "sawtooth":
        pattern = np.linspace(-scale, scale, pattern_length)
        return np.concatenate([np.repeat(pattern, periods), pattern[:remainder]])


@require_extension
def generate_dataframe(minutes=2, freq=100):

    start = datetime.datetime(2017, 1, 1, 0, 0)
    end = start + datetime.timedelta(minutes=minutes)

    # index
    ind = pd.date_range(start=start, end=end, freq="10000u", closed="left")
    n = len(ind)

    # dataframe
    return pd.DataFrame(
        {
            "random-walk": generate_data(n, func="walk"),
            "sin": generate_data(n, func="sin"),
            "sawtooth": generate_data(n, func="sawtooth"),
            "square": generate_data(n, func="square"),
        },
        index=ind,
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from builtins import dict, object
from future.utils import raise_from

import base64
import json

import boto3
import requests
from jose import jwk, jwt
from jose.utils import base64url_decode
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from requests.packages.urllib3.util.retry import Retry

# pennsieve
from pennsieve import log
from pennsieve.models import User


class UnauthorizedException(Exception):
    pass


class PennsieveRequest(object):
    def __init__(self, func, uri, *args, **kwargs):
        self._func = func
        self._uri = uri
        self._args = args
        self._kwargs = kwargs
        self._response = None

        self._logger = log.get_logger("pennsieve.base.PennsieveRequest")

    def raise_for_status(self, resp):
        try:
            resp.raise_for_status()
        except HTTPError as e:  # raise for status raise an HTTPError, so we can use it to grab the message
            if resp.text:
                raise_from(HTTPError(resp.content, response=resp), e)
            else:
                raise e
        return

    def _handle_response(self, resp):
        self._logger.debug(u"resp = {}".format(resp))
        self._logger.debug(u"resp.content = {}".format(resp.text))  # decoded unicode
        if resp.status_code in [requests.codes.forbidden, requests.codes.unauthorized]:
            raise UnauthorizedException()

        if not resp.status_code in [requests.codes.ok, requests.codes.created]:
            self.raise_for_status(resp)
        try:
            # return object from json
            resp.data = json.loads(resp.text)
        except:
            # if not json, still return response content
            resp.data = resp.text

    def call(self, timeout=None):
        self._response = self._func(
            self._uri, *self._args, timeout=timeout, **self._kwargs
        )
        self._handle_response(self._response)
        return self._response


class ClientSession(object):
    def __init__(self, settings):
        self._host = settings.api_host
        self._api_token = settings.api_token
        self._api_secret = settings.api_secret
        self._jwt = settings.jwt
        self._headers = settings.headers
        self._model_service_host = settings.model_service_host
        self._logger = log.get_logger("pennsieve.base.ClientSession")

        self._session = None
        self._token = None
        self._secret = None
        self._context = None
        self._organization = None
        self.profile = None
        self.settings = settings

    def authenticate(self, organization=None):
        """"""
        if self._jwt is None:
            self._authenticate_with_cognito(organization=organization)
        else:
            self._authenticate_with_jwt()

    def _authenticate_with_cognito(self, organization=None):
        """
        An API token is used to authenticate against the Pennsieve platform.
        The token that is returned from the API call will be used for all
        subsequent API calls.
        """

        cognito_config = self._get("/authentication/cognito-config")
        cognito_client_application_id = cognito_config["tokenPool"]["appClientId"]
        cognito_region_name = cognito_config["region"]

        # Authenticate to AWS Cognito
        #
        # Hack: stub the access and secret keys with empty values so boto does
        # not look for AWS credentials in the environment. Some versions of boto
        # fail when they cannot find AWS credentials even though Cognito does
        # not need creds.
        cognito_idp_client = boto3.client(
            "cognito-idp",
            region_name=cognito_region_name,
            aws_access_key_id="",
            aws_secret_access_key="",
        )
        response = cognito_idp_client.initiate_auth(
            AuthFlow="USER_PASSWORD_AUTH",
            AuthParameters={"USERNAME": self._api_token, "PASSWORD": self._api_secret},
            ClientId=cognito_client_application_id,
        )

        # Grab the tokens
        access_token_jwt = response["AuthenticationResult"]["AccessToken"]
        id_token_jwt = response["AuthenticationResult"]["IdToken"]

        # Since we passed the verification, we can now safely use the claims
        claims = jwt.get_unverified_claims(id_token_jwt)

        # Ensures that `self._session` exists
        self.session

        # Parse response, set session access token
        self.token = access_token_jwt
        self.profile = User.from_dict(self._get("/user/"))

        if organization is None:
            organization = claims["custom:organization_node_id"]

        self._set_org_context(organization)

    def _authenticate_with_jwt(self):
        """
        Use a JWT to make all subsequent requests to API.
        """
        # Force session creation:
        self.session
        self.token = self._jwt
        # Get the associated user and set the org context to the preferred
        # org set for the user. Additionally, if authentication fails at this
        # step don't attempt to reauthenticate recursively.
        user = self._get("/user", reauthenticate=False)
        organization = user["preferredOrganization"]
        self._set_org_context(organization)

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, value):
        self._token = value
        self._set_auth(value)

    def _set_org_context(self, organization_id):
        self._organization = organization_id
        self._session.headers["X-ORGANIZATION-ID"] = organization_id

    def _set_auth(self, session_token):
        self._session.headers["Authorization"] = "Bearer {}".format(session_token)

    @property
    def session(self):
        """
        Make requests-futures work within threaded/distributed environment.
        """
        if self._session is None:
            self._session = Session()
            self._set_auth(self.token)

            # Set global headers
            if self._headers:
                self._session.headers.update(self._headers)

            # Enable retries via urllib
            adapter = HTTPAdapter(
                max_retries=Retry(
                    total=self.settings.max_request_timeout_retries,
                    backoff_factor=0.5,
                    status_forcelist=[
                        500,
                        502,
                        503,
                        504,
                    ],  # Retriable errors (but not POSTs)
                )
            )
            self._session.mount("http://", adapter)
            self._session.mount("https://", adapter)

        return self._session

    def _make_request(self, func, uri, *args, **kwargs):
        self._logger.debug("~" * 60)
        self._logger.debug("uri = {} {}".format(func.__func__.__name__, uri))
        self._logger.debug("args = {}".format(args))
        self._logger.debug("kwargs = {}".format(kwargs))
        self._logger.debug("headers = {}".format(self.session.headers))
        return PennsieveRequest(func, uri, *args, **kwargs)

    def _call(self, method, endpoint, base="", reauthenticate=True, *args, **kwargs):
        if method == "get":
            func = self.session.get
        elif method == "put":
            func = self.session.put
        elif method == "post":
            func = self.session.post
        elif method == "delete":
            func = self.session.delete

        # serialize data
        if "data" in kwargs:
            kwargs["data"] = json.dumps(kwargs["data"])

        # we might specify a different host
        if "host" in kwargs:
            host = kwargs["host"]
            kwargs.pop("host")
        else:
            host = self._host

        # call endpoint
        uri = self._uri(endpoint, base=base, host=host)
        req = self._make_request(func, uri, *args, **kwargs)
        resp = self._get_response(req, reauthenticate=reauthenticate)

        return resp.data

    def _uri(self, endpoint, base, host=None):
        if host is None:
            host = self._host
        return "{}{}{}".format(host, base, endpoint)

    def _get(self, endpoint, *args, **kwargs):
        return self._call("get", endpoint, *args, **kwargs)

    def _post(self, endpoint, *args, **kwargs):
        return self._call("post", endpoint, *args, **kwargs)

    def _put(self, endpoint, *args, **kwargs):
        return self._call("put", endpoint, *args, **kwargs)

    def _del(self, endpoint, *args, **kwargs):
        return self._call("delete", endpoint, *args, **kwargs)

    def _get_response(self, req, reauthenticate=True):
        try:
            return req.call(timeout=self.settings.max_request_time)
        except UnauthorizedException as e:
            if self._token is None or reauthenticate is False:
                raise e

            # try to refresh the session and re-request
            self.authenticate(self._organization)
            return req.call(timeout=self.settings.max_request_time)

    def register(self, *components):
        """
        Register API component with session. Components should all be of
        APIBase type and have a name and base_uri property.

        The registered component will have reference to base session to
        make higher-level calls outside of its own scope, if needed.
        """
        # initialize
        for component in components:
            c = component(session=self)
            assert len(component.name) > 1, "Invalid API component name"
            # component is accessible via session.(name)
            self.__dict__.update({component.name: c})

    @property
    def headers(self):
        return self.session.headers



"""
Description
  The profile management system is used to create multiple profiles for
  different Pennsieve accounts or organizations. The user can also set/unset
  global settings which apply to all profiles.

  Settings are loaded in the following order, with each tier overriding the
  last:

    1. Global settings are loaded first
    2. Default profile, if configured
    3. Command line arguments, e.g.: --profile=<name>
    4. Environment variables, e.g.: PENNSIEVE_API_TOKEN

Usage:
  pennsieve-profile create [<name>]
  pennsieve-profile show [<name>]
  pennsieve-profile delete [<name>] [-f | --force]
  pennsieve-profile list [-c | --contents]
  pennsieve-profile set-default [<name>]
  pennsieve-profile unset-default [-f | --force]
  pennsieve-profile status
  pennsieve-profile set <key> <value> [-f | --force] [--profile=<name>]
  pennsieve-profile unset <key> [-f | --force] [--profile=<name>]
  pennsieve-profile keys [--profile=<name>]
  pennsieve-profile version
  pennsieve-profile help

Options:
  -c --contents       List profiles with contents. Also lists global settings, if any
  -f --force          Attempt action without prompting for confirmation. Use with care
  --profile=<name>    Use specified profile (instead of default)
  -h --help           Show help

Advanced commands:
  set (global|<name>) <key> <value>   Set key/value pair for given profile, or globally
  unset (global|<name>) <key>         Unset key/value pair for given profile, or globally
  keys                                List all available keys and their default values

For additional features, install the Pennsieve CLI Agent:
https://developer.pennsieve.io/agent
"""

from __future__ import absolute_import, print_function
from builtins import input

import io
import os

from docopt import docopt

import pennsieve
from pennsieve import DEFAULT_SETTINGS, Pennsieve, Settings


def main():
    args = docopt(__doc__, version=pennsieve.__version__)

    # Test for these two commands first as they
    # do not require a Pennsieve client or reading the settings file
    if args["help"]:
        print(__doc__.strip("\n"))
        return
    elif args["version"]:
        print(pennsieve.__version__)
        return

    settings = Settings(args["--profile"])
    if not os.path.exists(settings.config_file):
        setup_assistant(settings)
    elif args["create"]:
        create_profile(settings, args["<name>"])
    elif args["show"]:
        show_profile(settings, args["<name>"])
    elif args["delete"]:
        delete_profile(settings, args["<name>"], args["--force"])
    elif args["list"]:
        list_profiles(settings, args["--contents"])
    elif args["set-default"]:
        set_default(settings, args["<name>"])
    elif args["unset-default"]:
        unset_default(settings, args["--force"])
    elif args["set"]:
        set_key(
            settings, args["<key>"], args["<value>"], args["--profile"], args["--force"]
        )
    elif args["unset"]:
        unset_key(settings, args["<key>"], args["--profile"], args["--force"])
    elif args["keys"]:
        list_keys(settings, args["--profile"])
    elif args["status"]:
        ps = Pennsieve(args["--profile"])
        show_status(ps)
    else:
        invalid_usage()

    with io.open(settings.config_file, "w") as configfile:
        settings.config.write(configfile)


def setup_assistant(settings):
    settings.config.clear()

    print("Pennsieve profile setup assistant")

    settings.config["global"] = {"default_profile": "none"}

    print("Create a profile:")
    create_profile(settings)

    print(
        "Setup complete. Run 'pennsieve-profile help' for available commands and actions"
    )


# User commands
# =======================================================


def create_profile(settings, name=None):
    if not name:
        name = input("  Profile name [default]: ") or "default"

    if name in settings.config:
        if name in ("global", "agent", "none"):
            print(
                "Profile name '{}' reserved for system. Please try a different name".format(
                    name
                )
            )
        else:
            print("Profile '{}' already exists".format(name))
        abort()

    print("Creating profile '{}'".format(name))
    settings.config[name] = {}

    settings.config[name]["api_token"] = input("  API token: ")
    settings.config[name]["api_secret"] = input("  API secret: ")

    if settings.config["global"]["default_profile"] == "none":
        settings.config["global"]["default_profile"] = name
        print("Default profile: {}".format(name))

    else:
        if yesno_prompt("Would you like to set '{}' as default (Y/n)? ".format(name)):
            set_default(settings, [name])


def delete_profile(settings, name, force):
    if not name:
        name = input("  Profile to delete: ")

    if not valid_name(settings, name):
        abort()

    if not force:
        if not yesno_prompt("Delete profile '{}' (Y/n)? ".format(name)):
            print("abort")
            abort()

    print("Deleting profile '{}'".format(name))
    settings.config.remove_section(name)

    if settings.config["global"]["default_profile"] == name:
        settings.config["global"]["default_profile"] = "none"
        print(
            "\033[31m* Warning: default profile unset. Use 'ps profile set-default <name>' to set a new default\033[0m"
        )


def list_profiles(settings, contents):
    """
    Lists all profiles
    """
    print("Profiles:")
    for section in settings.config.sections():
        if section not in ["global"]:
            if section == settings.config["global"]["default_profile"]:
                print("* \033[32m{}\033[0m".format(section))
            else:
                print("  {}".format(section))
            if contents:
                print_profile(settings, section, 4)
    if contents:
        if len(settings.config["global"]) > 1:
            print("Global Settings:")
            print_profile(settings, "global", 2)


def set_default(settings, name):
    if not name:
        name = input("  Profile name to set as default: ")

    if not valid_name(settings, name):
        abort()

    print("Default profile: {}".format(name))
    settings.config["global"]["default_profile"] = name


def unset_default(settings, force):
    original = settings.config["global"]["default_profile"]

    if not force:
        if not yesno_prompt("Unset default profile '{}' (Y/n)? ".format(original)):
            print("abort")
            abort()

    print(
        "Default profile '{}' unset. Using global settings and environment variables".format(
            original
        )
    )
    settings.config["global"]["default_profile"] = "none"


def show_profile(settings, profile):
    if profile:
        name = profile
    else:
        name = settings.default_profile

    if name == "global" or valid_name(settings, name):
        print("{} contents:".format(name))
        print_profile(settings, name, 2, True)


# Advanced commands
# =======================================================


def set_key(settings, key, value, profile, force):
    if not key in DEFAULT_SETTINGS:
        print(
            "Invalid key: '{}'\n see 'ps profile keys' for available keys".format(key)
        )
        return

    if profile:
        name = profile
    else:
        name = settings.default_profile

    if not name in settings.config:
        print("Profile '{}' does not exist".format(name))
        return

    if not force and key in settings.config[name]:
        if not yesno_prompt("{}: {} already set. Overwrite (Y/n)? ".format(name, key)):
            abort()

    print("{}: {}={}".format(name, key, value))
    settings.config[name][key] = value


def unset_key(settings, key, profile, force):
    if profile:
        name = profile
    else:
        name = settings.default_profile

    if not name in settings.config:
        print("Profile '{}' does not exist".format(name))
        abort()

    if not key in settings.config[name]:
        print("{}: {} not set".format(name, key))
        abort()

    if not force and key in settings.config[name]:
        if not yesno_prompt("{}: Unset {} (Y/n)? ".format(name, key)):
            abort()

    print("{}: {} unset".format(name, key))
    settings.config[name].pop(key)


def list_keys(settings, profile):
    if profile:
        name = profile
    else:
        name = settings.default_profile

    print("Keys and default values for '{}'".format(name))
    for key, value in sorted(settings.profiles[name].items()):
        print("  {} : {}".format(key, value))


def show_status(ps):
    print("Active profile:\n  \033[32m{}\033[0m\n".format(ps.settings.active_profile))

    if ps.settings.env:
        key_len = 0
        value_len = 0
        for key, value in ps.settings.env.items():
            valstr = "{}".format(value)
            key_len = max(key_len, len(key))
            value_len = max(value_len, len(valstr))

        print("Environment variables:")
        print(
            "  \033[4m{:{key_len}}    {:{value_len}}    {}".format(
                "Key",
                "Value",
                "Environment Variable\033[0m",
                key_len=key_len,
                value_len=value_len,
            )
        )
        for value, evar in sorted(ps.settings.env.items()):
            valstr = "{}".format(value)
            # get internal variable
            ivar = ""
            for k, v in ps.settings.__dict__.items():
                if str(v) == str(evar):
                    ivar = k
            print(
                "  {:{key_len}}    {:{value_len}}    {}".format(
                    ivar, evar, valstr, key_len=key_len, value_len=value_len
                )
            )
        print()

    print("Pennsieve environment:")
    print("  User               : {}".format(ps.profile.email))
    print("  Organization       : {} (id: {})".format(ps.context.name, ps.context.id))
    print("  API Location       : {}".format(ps.settings.api_host))


# Helper functions
# =======================================================


def abort():
    exit(1)


def yesno_prompt(msg):
    return input(msg).lower() in ("y", "yes")


def invalid_usage():
    print("Invalid usage. See `pennsieve-profile help` for available commands")
    abort()


def valid_name(settings, name):
    if name not in settings.config or name == "global":
        print("Profile '{}' does not exist".format(name))
        return False
    return True


def print_profile(settings, name, indent=0, show_all=False):
    if show_all:
        key_len = 0
        for key in settings.profiles[name].keys():
            key_len = max(key_len, len(key))

        for key, value in sorted(settings.profiles[name].items()):
            if key != "default_profile":
                if key in settings.config[name] and name != "global":
                    print(
                        " " * indent
                        + "{:{key_len}} : \033[32m{}\033[0m ({})".format(
                            key, value, name, key_len=key_len
                        )
                    )
                elif key in settings.config["global"]:
                    print(
                        " " * indent
                        + "{:{key_len}} : \033[34m{}\033[0m (global)".format(
                            key, value, key_len=key_len
                        )
                    )
                else:
                    print(
                        " " * indent
                        + "{:{key_len}} : \033[0m{}\033[0m".format(
                            key, value, key_len=key_len
                        )
                    )

    else:
        key_len = 0
        for key in settings.config[name].keys():
            key_len = max(key_len, len(key))

        for key, value in sorted(settings.config[name].items()):
            if key != "default_profile":
                print(
                    " " * indent
                    + "{:{key_len}} : {}".format(key, value, key_len=key_len)
                )



import time
import uuid

import pytest
from pkg_resources import resource_filename

from pennsieve.api.agent import AgentError
from pennsieve.models import TimeSeries


def _resource_path(fname):
    return resource_filename("tests.resources", fname)


# All test assets need to use this "test-78f3ea50" prefix so failures caused by
# deleting datasets before processing all assests can be filtered out of
# production error notifications.
FILE1 = _resource_path("test-78f3ea50.txt")
FILE2 = _resource_path("test-78f3ea50.csv")
FILE3 = _resource_path("test-78f3ea50.png")
FILE4 = _resource_path("test-78f3ea50-b.txt")
FILE_EMPTY = _resource_path("test-78f3ea50.empty")
FLAT_DIR = _resource_path("flat_dir")
NESTED_DIR = _resource_path("nested_dir")
INNER_DIR = "inner_dir"


@pytest.mark.agent
@pytest.mark.parametrize(
    "upload_args,n_files",
    [
        ([FILE4], 1),
    ],
)  # Single file
def test_get_by_filename(dataset, upload_args, n_files):

    dataset.upload(*upload_args)
    dataset.update()

    for i in range(10):
        pp = dataset.get_items_by_name("test-78f3ea50-b.txt")
        if len(pp) > 0 and pp[0].state == "READY":
            break
        else:
            time.sleep(0.5)

    packages = dataset.get_packages_by_filename("test-78f3ea50-b.txt")
    assert len(packages) == 1
    assert packages[0].name == "test-78f3ea50-b.txt"


@pytest.mark.agent
@pytest.mark.parametrize(
    "upload_args,n_files",
    [
        ([FILE1], 1),
        ([[FILE1, FILE2]], 2),
    ],  # Single file  # Multiple files
)
def test_upload_to_collection(dataset, upload_args, n_files):
    collection = dataset.create_collection(str(uuid.uuid4()))
    collection.upload(*upload_args)

    assert len(collection.items) == n_files


@pytest.mark.agent
@pytest.mark.parametrize(
    "upload_args, n_files",
    [
        ([FLAT_DIR], 3),
        ([NESTED_DIR], 1),
        ([NESTED_DIR + "/" + INNER_DIR], 2),
    ],
)
def test_upload_directory(dataset, upload_args, n_files):
    collection = dataset.create_collection(str(uuid.uuid4()))
    collection.upload(upload_args)
    assert len(collection.items) == n_files


@pytest.mark.agent
def test_upload_recursive(dataset):
    collection = dataset.create_collection(str(uuid.uuid4()))
    collection.upload(NESTED_DIR, recursive=True)

    assert len(collection.items) == 1
    nested_dir = collection.items[0]
    assert len(nested_dir.items) == 2
    inner_dir = next(pkg for pkg in nested_dir.items if pkg.name == INNER_DIR)
    assert len(inner_dir.items) == 2


@pytest.mark.agent
def test_upload_recursive_flag_is_not_allowed_with_file(dataset):
    collection = dataset.create_collection(str(uuid.uuid4()))
    with pytest.raises(AgentError):
        collection.upload(FILE1, recursive=True)


@pytest.mark.agent
def test_upload_cannot_upload_multiple_directories(dataset):
    with pytest.raises(AgentError):
        dataset.upload(FLAT_DIR, FILE1)


@pytest.mark.agent
@pytest.mark.parametrize(
    "upload_args,n_files",
    [
        ([FILE1], 1),  # Single file
        ([FILE1, FILE2], 2),  # Multiple files, separate arguments
        ([[FILE1, FILE2]], 2),  # Multiple files, single argument list
    ],
)
def test_upload_to_dataset(dataset, upload_args, n_files):
    """
    Note: ETL will fail since destination will likely be removed
          before being processed.
    """
    c = len(dataset.items)
    dataset.upload(*upload_args)
    dataset.update()
    assert len(dataset.items) == c + n_files


@pytest.mark.agent
@pytest.mark.parametrize(
    "append_args,n_files",
    [
        ([FILE1], 1),  # Single file
        ([FILE1, FILE2], 2),  # Multiple files, separate arguments
        ([[FILE1, FILE2]], 2),  # Multiple files, single argument list
    ],
)
def test_append(dataset, append_args, n_files):
    """
    Note: ETL will fail for append, because... it's a text file.
          But also because the destination will likely be removed
          before the ETL actually places the new node there.
    """
    # TimeSeries package to append into...
    pkg = TimeSeries("Rando Timeseries")
    dataset.add(pkg)

    # upload/append file into package
    pkg.append_files(*append_args)
    # TODO: assert append was successful


@pytest.mark.agent
def test_progress_for_empty_files(dataset):
    dataset.upload(FILE_EMPTY, display_progress=True)



import datetime
from collections import namedtuple

import pytest
from past.builtins import unicode  # Alias of str in Python 3

from pennsieve.models import (
    DataPackage,
    ModelProperty,
    ModelPropertyEnumType,
    ModelPropertyType,
)
from tests.utils import create_test_dataset, current_ts, get_test_client


@pytest.mark.parametrize(
    "from_type,pennsieve_type,data_type",
    [
        (int, "long", int),
        (ModelPropertyType(data_type=int), "long", int),
        (ModelPropertyType(data_type=unicode), "string", unicode),
        (datetime.date, "date", datetime.date),
        ("date", "date", datetime.datetime),
        (ModelPropertyType(data_type=str, format="date"), "string", str),
    ],
)
def test_model_type_conversion(from_type, pennsieve_type, data_type):
    property_type = ModelPropertyType._build_from(from_type)
    assert property_type.data_type == data_type
    assert property_type._pennsieve_type == pennsieve_type


@pytest.mark.parametrize(
    "json,data_type,format,unit",
    [
        ("string", unicode, None, None),
        ({"type": "string", "format": "date"}, unicode, "date", None),
        ({"type": "double", "unit": "kg"}, float, None, "kg"),
    ],
)
def test_model_type_from_dict(json, data_type, format, unit):
    decoded = ModelPropertyType._build_from(json)
    assert isinstance(decoded, ModelPropertyType)
    assert decoded.data_type == data_type
    assert decoded.format == format
    assert decoded.unit == unit


def test_rollback_model_creation_with_invalid_properties(dataset):
    model_name = "New_Model_{}".format(current_ts())
    invalid_schema = [("an_integer", int, "An Integer")]

    # Creating the model succeeds, but then creating properties fails
    with pytest.raises(Exception) as e:
        dataset.create_model(
            model_name,
            schema=invalid_schema,
        )
    assert "Could not create model properties" in str(e.value)

    # Model creation is rolled back
    assert model_name not in dataset.models()


def test_date_formatting():
    d1 = datetime.datetime(2018, 8, 24, 15, 11, 25)
    assert (
        ModelPropertyType(datetime.datetime)._encode_value(d1)
        == "2018-08-24T15:11:25.000000+00:00"
    )

    d2 = datetime.datetime(2018, 8, 24, 15, 11, 25, 1)
    assert (
        ModelPropertyType(datetime.datetime)._encode_value(d2)
        == "2018-08-24T15:11:25.000001+00:00"
    )


def test_boolean_string():
    assert ModelPropertyType(bool)._decode_value("false") == False
    assert ModelPropertyType(bool)._decode_value("true") == True
    assert ModelPropertyType(bool)._decode_value("nope") == True


def test_models(dataset):
    schema = [
        ("an_integer", int, "An Integer", True),
        ("a_bool", bool),
        ("a_string", str),
        ("a_datetime", datetime.datetime),
    ]
    display_name = "A New Property"
    description = "a new description"
    values = {
        "an_integer": 100,
        "a_bool": True,
        "a_string": "fnsdlkn#$#42nlfds$3nlds$#@$23fdsnfkls",
        "a_datetime": datetime.datetime.now(),
    }

    #################################
    ## Models
    ################################

    models = dataset.models()

    new_model = dataset.create_model(
        "New_Model_{}".format(current_ts()), "A New Model", "a new model", schema
    )

    assert len(dataset.models()) == len(models) + 1

    assert dataset.get_model(new_model.id) == new_model
    assert dataset.get_model(new_model.type) == new_model

    # Check that local changes get propagated
    new_model.add_property("a_new_property", str, display_name)
    new_model.description = description
    new_model.update()
    new_model = dataset.get_model(new_model.id)
    assert new_model.description == description
    assert new_model.get_property("a_new_property").display_name == display_name

    new_model.add_properties(
        [
            ("a_new_float", float),
            {"name": "a_new_int", "data_type": int},
            "a_new_string",
        ]
    )
    assert new_model.get_property("a_new_float").type == float
    assert new_model.get_property("a_new_int").type == int
    assert new_model.get_property("a_new_string").type == unicode

    nc_one = new_model.create_record(values)
    nc_two = new_model.create_record(
        {
            "an_integer": 1,
            "a_bool": False,
            "a_string": "",
            "a_datetime": datetime.datetime.now(),
        }
    )
    nc_three = new_model.create_record(
        {
            "an_integer": 10000,
            "a_bool": False,
            "a_string": "43132312",
            "a_datetime": datetime.datetime.now(),
        }
    )
    nc_four = new_model.create_record(
        {"an_integer": 9292, "a_datetime": datetime.datetime.now()}
    )

    nc_delete_one = new_model.create_record(
        {"an_integer": 28, "a_datetime": datetime.datetime.now()}
    )
    nc_delete_two = new_model.create_record(
        {"an_integer": 300, "a_datetime": datetime.datetime.now()}
    )

    with pytest.raises(Exception):
        new_model.create_record()

    new_models_old = new_model.get_all()
    assert new_model.get_all(limit=1) == new_models_old[:1]
    assert new_model.get_all(limit=2, offset=2) == new_models_old[2:4]

    new_model.delete_records(nc_delete_one, nc_delete_two.id)
    new_models = new_model.get_all()

    assert len(new_models) == (len(new_models_old) - 2)

    assert nc_two.model == new_model
    assert nc_two.get("a_string") == ""

    nc_four.set("a_string", "hello")
    assert nc_four.get("a_string") == new_model.get(nc_four).get("a_string")

    with pytest.raises(Exception):
        nc_four.set("an_integer", datetime.datetime.now())

    assert nc_four.get("an_integer") == 9292
    nc_four.set("an_integer", 10)
    assert nc_four.get("an_integer") == 10

    nc_delete_three = new_model.create_record(
        {"an_integer": 684, "a_string": "delete me"}
    )
    assert len(new_model.get_all()) == len(new_models) + 1
    nc_delete_three.delete()
    assert len(new_model.get_all()) == len(new_models)

    # cannot add a record id column using an existing name
    with pytest.raises(ValueError):
        new_model.get_all().as_dataframe(
            record_id_column_name=list(new_model.get_all().type.schema.keys())[0]
        )

    # assert no extra columns are added by default
    df_cs_no_rec = new_model.get_all().as_dataframe()
    assert len(df_cs_no_rec.columns) == len(new_model.get_all().type.schema.keys())

    # assert record id column is added when arg is present and valid
    df_cs = new_model.get_all().as_dataframe(record_id_column_name="record_id")

    # confirm that all record ids are present in this dataframe
    assert "record_id" in df_cs.columns
    for record in new_model.get_all():
        assert not df_cs.query("record_id == @record.id").empty

    #################################
    ## Relationships
    ################################

    relationships = dataset.relationships()

    new_relationship = dataset.create_relationship_type(
        "New_Relationship_{}".format(current_ts()), "a new relationship"
    )

    assert len(dataset.relationships()) == len(relationships) + 1

    assert dataset.get_relationship(new_relationship.id) == new_relationship
    assert dataset.get_relationship(new_relationship.type) == new_relationship

    nr_one = new_relationship.relate(nc_one, nc_two)
    nr_four = new_relationship.relate(nc_four, nc_one)
    nr_five = new_relationship.relate(nc_four, nc_two)

    nr_two = nc_two.relate_to(nc_three, new_relationship)
    nr_three = nc_three.relate_to(nc_four, new_relationship)
    nr_six = nc_four.relate_to(nc_three, new_relationship)
    nr_seven = nc_four.relate_to(nc_one, relationship_type="goes_to")

    assert nr_seven[0].destination == nc_one.id
    assert nr_seven[0].source == nc_four.id
    assert nr_seven[0].type == "goes_to"
    assert len(nc_four.get_related(new_model.type)) == 5

    new_relationships = new_relationship.get_all()

    nr_delete_three = new_relationship.relate(nc_one, nc_three)

    assert len(new_relationship.get_all()) == len(new_relationships) + 1
    nr_delete_three.delete()
    assert len(new_relationship.get_all()) == len(new_relationships)

    df_rs = new_relationship.get_all().as_dataframe()

    p = DataPackage("test-csv", package_type="CSV")
    dataset.add(p)
    dataset.update()
    assert p.exists

    p.relate_to(nc_one)
    p.relate_to(nc_two)
    nc_three.relate_to(p, new_relationship)
    new_relationship.relate(nc_four, p)

    assert len(nc_four.get_related(new_model.type)) == 5


def test_simple_model_properties(dataset):

    # Define properties as tuples
    model_with_basic_props_1 = dataset.create_model(
        "Basic_Props_1",
        description="a new description",
        schema=[
            ("an_integer", int, "An Integer", True),
            ("a_bool", bool),
            ("a_string", str),
            ("a_date", datetime.datetime),
        ],
    )

    assert dataset.get_model(model_with_basic_props_1.id) == model_with_basic_props_1

    # Add a property with a description
    model_with_basic_props_1.add_property(
        "a_new_property", float, display_name="Weight", description="some metric"
    )

    updated_model_1 = dataset.get_model(model_with_basic_props_1.id)

    test_prop = updated_model_1.get_property("a_new_property")
    assert test_prop.display_name == "Weight"
    assert test_prop.description == "some metric"

    # Define properties as ModelProperty objects
    model_with_basic_props_2 = dataset.create_model(
        "Basic_Props_2",
        description="a new description",
        schema=[
            ModelProperty("name", data_type=str, title=True, required=True),
            ModelProperty("age", data_type=int),
            ModelProperty("DOB", data_type=datetime.datetime),
        ],
    )

    assert dataset.get_model(model_with_basic_props_2.id) == model_with_basic_props_2

    # Add a property
    model_with_basic_props_2.add_property(
        "weight2", ModelPropertyType(data_type=float), display_name="Weight"
    )

    updated_model_2 = dataset.get_model(model_with_basic_props_2.id)

    assert updated_model_2.get_property("weight2").display_name == "Weight"
    assert updated_model_2.get_property("name").required == True

    # Define properties as ModelProperty objects with ModelPropertyType data_type
    model_with_basic_props_3 = dataset.create_model(
        "Basic_Props_3",
        description="a new description",
        schema=[
            ModelProperty(
                "name", data_type=ModelPropertyType(data_type=str), title=True
            ),
            ModelProperty("age", data_type=ModelPropertyType(data_type=int)),
            ModelProperty("DOB", data_type=ModelPropertyType(data_type=str)),
        ],
    )

    assert dataset.get_model(model_with_basic_props_3.id) == model_with_basic_props_3

    # Add a property
    model_with_basic_props_3.add_property(
        "weight3", ModelPropertyType(data_type=float), display_name="Weight"
    )

    updated_model_3 = dataset.get_model(model_with_basic_props_3.id)

    assert updated_model_3.get_property("weight3").display_name == "Weight"

    # Reverse look up property data types
    ModelProperty("name", data_type="string", title=True, required=True)
    model_with_basic_props_4 = dataset.create_model(
        "Basic_Props_4",
        description="a new description",
        schema=[
            ModelProperty("name", data_type="string", title=True, required=True),
            ModelProperty("age", data_type="long"),
            ModelProperty("DOB", data_type="date"),
        ],
    )

    assert dataset.get_model(model_with_basic_props_4.id) == model_with_basic_props_4

    # Add a property
    model_with_basic_props_4.add_property(
        "weight4", ModelPropertyType(data_type="double"), display_name="Weight"
    )

    updated_model_4 = dataset.get_model(model_with_basic_props_4.id)

    assert updated_model_4.get_property("weight4").display_name == "Weight"
    assert updated_model_4.get_property("name").required == True


def test_model_property_default_field_inherits_from_required_field(dataset):
    """
    Unless explictly set, the `default` field on a property takes the value of
    the `required` field.
    """
    model = dataset.create_model(
        "Default_properties_model",
        description="",
        schema=[
            ModelProperty("name", data_type="string", title=True),
            ModelProperty("required", data_type="string", required=True),
            ModelProperty("not_required", data_type="string", required=False),
        ],
    )

    model = dataset.get_model(model)
    assert model.get_property("required").required is True
    assert model.get_property("required").default is True

    assert model.get_property("not_required").required is False
    assert model.get_property("not_required").default is False

    model.create_record({"name": "Record #1", "required": "yes"})


def test_model_title_is_a_required_property_by_default(dataset):
    model = dataset.create_model(
        "Concept_title_model",
        description="",
        schema=[
            ModelProperty("name", data_type="string", title=True),
        ],
    )

    model = dataset.get_model(model)
    assert model.get_property("name").title is True
    assert model.get_property("name").required is True


def test_complex_model_properties(dataset):
    model_with_complex_props = dataset.create_model(
        "Complex_Props",
        description="a new description",
        schema=[
            ModelProperty(
                "name", data_type=ModelPropertyType(data_type=str), title=True
            ),
            ModelProperty("age", data_type=ModelPropertyType(data_type=int)),
            ModelProperty(
                "email", data_type=ModelPropertyType(data_type=str, format="email")
            ),
        ],
    )

    assert dataset.get_model(model_with_complex_props.id) == model_with_complex_props

    # Add a property
    model_with_complex_props.add_property(
        "weight", ModelPropertyType(data_type=float, unit="kg"), display_name="Weight"
    )

    updated_model = dataset.get_model(model_with_complex_props.id)

    weight_property = updated_model.get_property("weight")
    assert weight_property.display_name == "Weight"
    assert weight_property.type == float
    assert weight_property._type.data_type == float
    assert weight_property._type.format == None
    assert weight_property.unit == "kg"

    email_property = model_with_complex_props.get_property("email")
    assert email_property.type == unicode
    assert email_property._type.data_type == unicode
    assert email_property._type.format.lower() == "email"
    assert email_property.unit == None

    good_values = {"name": "Bob", "age": 1, "email": "test@test.com", "weight": 10}
    good_record = updated_model.create_record(good_values)

    bad_values = {"name": "Bob", "age": 1, "email": "123455", "weight": 10}

    with pytest.raises(Exception):
        bad_record = updated_model.create_record(bad_values)


def test_get_connected(dataset):
    model_1 = dataset.create_model(
        "Model_A",
        description="model a",
        schema=[
            ModelProperty(
                "prop1", data_type=ModelPropertyType(data_type=str), title=True
            )
        ],
    )

    related_models = model_1.get_connected()
    # For a single, unconnected model, it should return nothing
    assert len(related_models) == 0

    model_2 = dataset.create_model(
        "Model_B",
        description="model b",
        schema=[
            ModelProperty(
                "prop1", data_type=ModelPropertyType(data_type=str), title=True
            )
        ],
    )

    relationship = dataset.create_relationship_type(
        "New_Relationship_{}".format(current_ts()), "a new relationship"
    )

    model_instance_1 = model_1.create_record({"prop1": "val1"})
    model_instance_2 = model_2.create_record({"prop1": "val1"})
    model_instance_1.relate_to(model_instance_2, relationship)

    related_models = model_1.get_connected()
    assert len(related_models) == 2

    # For a connected model, return all connections + the model itself
    related_models = model_2.get_connected()
    assert len(related_models) == 2

    # Check that get_connected_models from the dataset object also works
    related_models = dataset.get_connected_models(model_1.id)
    assert len(related_models) == 2

    related_models = dataset.get_connected_models(model_2.id)
    assert len(related_models) == 2


def test_array_model_properties(dataset):
    model = dataset.create_model(
        "Array_Properties",
        description="Description",
        schema=[
            ModelProperty("name", data_type=str, title=True),
            ModelProperty(
                "int_array",
                data_type=ModelPropertyEnumType(data_type=int, multi_select=True),
            ),
        ],
    )

    int_array = model.get_property("int_array")
    assert int_array.type == int
    assert int_array.multi_select == True
    assert int_array.enum == None

    record = model.create_record({"name": "A", "int_array": [1, 2, 3]})
    bulk_records = model.create_records(
        [
            {"name": "B", "int_array": [1, 2, 3]},
            {"name": "B", "int_array": [1, 2, 3]},
            {"name": "B", "int_array": [1, 2, 3]},
        ]
    )

    gotten = model.get_all()
    assert all(x.values["int_array"] == [1, 2, 3] for x in gotten)


def test_enum_model_properties(dataset):
    model = dataset.create_model(
        "Enum_Props",
        description="a new description",
        schema=[
            ModelProperty("name", data_type=str, title=True),
            ModelProperty(
                "int_enum",
                data_type=ModelPropertyEnumType(
                    data_type=float, enum=[1.0, 2.0, 3.0], unit="cm", multi_select=False
                ),
            ),
            ModelProperty(
                "str_enum",
                data_type=ModelPropertyEnumType(
                    data_type=str, enum=["foo", "bar", "baz"], multi_select=True
                ),
            ),
        ],
    )

    assert model == dataset.get_model(model.id)

    int_enum = model.get_property("int_enum")
    str_enum = model.get_property("str_enum")

    assert int_enum.type == float
    assert int_enum.multi_select == False
    assert int_enum.enum == [1.0, 2.0, 3.0]
    assert int_enum.unit == "cm"

    assert str_enum.type == unicode
    assert str_enum.multi_select == True
    assert str_enum.enum == ["foo", "bar", "baz"]

    record = model.create_record(
        {"name": "A", "int_enum": 1.0, "str_enum": ["foo", "bar"]}
    )

    gotten = model.get_all()[0]
    assert gotten.values["int_enum"] == 1.0
    assert set(gotten.values["str_enum"]) == set(["foo", "bar"])

    with pytest.raises(Exception, match=r"Value '5' is not a member*"):
        model.create_record({"name": "A", "int_enum": 5, "str_enum": ["foo", "bar"]})

    with pytest.raises(Exception, match=r"Value 'blah' is not a member*"):
        model.create_record({"name": "A", "int_enum": 1.0, "str_enum": ["blah"]})


Graph = namedtuple(
    "TestGraph",
    ["dataset", "models", "model_records", "relationships", "relationship_records"],
)


@pytest.fixture(scope="module")
def simple_graph(client):
    """
    Creates a small test graph in an independent dataset to de-couple
    from other tests
    """
    test_dataset = create_test_dataset(client)
    model_1 = test_dataset.create_model(
        "Model_A",
        description="model a",
        schema=[
            ModelProperty(
                "prop1", data_type=ModelPropertyType(data_type=str), title=True
            )
        ],
    )

    model_2 = test_dataset.create_model(
        "Model_B",
        description="model b",
        schema=[
            ModelProperty(
                "prop1", data_type=ModelPropertyType(data_type=str), title=True
            )
        ],
    )

    relationship = test_dataset.create_relationship_type(
        "New_Relationship_{}".format(current_ts()), "a new relationship"
    )

    model_instance_1 = model_1.create_record({"prop1": "val1"})
    model_instance_2 = model_2.create_record({"prop1": "val1"})
    model_instance_1.relate_to(model_instance_2, relationship)

    graph = Graph(
        test_dataset,
        models=[model_1, model_2],
        model_records=[model_instance_1, model_instance_2],
        relationships=[relationship],
        relationship_records=None,
    )
    yield graph

    ds_id = test_dataset.id
    client._api.datasets.delete(test_dataset)
    all_dataset_ids = [x.id for x in client.datasets()]
    assert ds_id not in all_dataset_ids
    assert not test_dataset.exists


def test_get_related_models(simple_graph):
    related_models = simple_graph.models[0].get_related()
    assert len(related_models) == 1
    assert related_models[0].type == simple_graph.models[1].type


def test_related_records_pagination(dataset):
    patient = dataset.create_model(
        "patient",
        description="patient",
        schema=[
            ModelProperty(
                "name", data_type=ModelPropertyType(data_type=str), title=True
            )
        ],
    )

    visit = dataset.create_model(
        "visit",
        description="visit",
        schema=[
            ModelProperty(
                "field", data_type=ModelPropertyType(data_type=str), title=True
            )
        ],
    )

    attends = dataset.create_relationship_type("attends", "an attendance")

    fred = patient.create_record({"name": "Fred"})
    visits = visit.create_records([{"field": str(i)} for i in range(200)])

    fred.relate_to(visits, attends)

    # Get all records
    gotten = fred.get_related()
    assert len(gotten) == 200
    assert [int(r.get("field")) for r in gotten] == list(range(200))


def test_stringified_boolean_values(dataset):
    ppatient = dataset.create_model(
        "potential_patient",
        description="potential patient",
        schema=[
            ModelProperty(
                "name", data_type=ModelPropertyType(data_type=str), title=True
            ),
            ModelProperty("sick", data_type=ModelPropertyType(data_type=bool)),
        ],
    )

    ppatient.create_record({"name": "Fred", "sick": "false"})
    ppatient.create_record({"name": "Joe", "sick": "true"})
    ppatient.create_record({"name": "Adele", "sick": "plop"})

    # Get all records
    gotten = ppatient.get_all()
    assert len(gotten) == 3
    assert gotten[0]._values["sick"].value == False
    assert gotten[1]._values["sick"].value == True
    assert gotten[2]._values["sick"].value == True


def test_get_topology(simple_graph):
    topology = simple_graph.dataset.get_topology()
    assert "models" in topology
    assert len(topology["models"]) == 2
    assert "relationships" in topology
    assert len(topology["relationships"]) == 1


def test_get_graph_summary(simple_graph):
    summary = simple_graph.dataset.get_graph_summary()
    expected_fields = [
        "modelCount",
        "modelRecordCount",
        "modelSummary",
        "relationshipCount",
        "relationshipRecordCount",
        "relationshipSummary",
        "relationshipTypeCount",
        "relationshipTypeSummary",
    ]
    for f in expected_fields:
        assert f in summary.keys()

    assert summary["modelCount"] == 2
    assert summary["modelRecordCount"] == 2
    assert summary["relationshipCount"] == 1
    assert summary["relationshipRecordCount"] == 1
    assert summary["relationshipTypeCount"] == 1


def test_simple_query(simple_graph):
    model_a = simple_graph.dataset.get_model("Model_A")
    model_b = simple_graph.dataset.get_model("Model_B")

    # Filter via a model name:
    assert len(model_a.query().filter("prop1", "eq", "val1").run()) == 1
    assert len(model_a.query().filter("prop1", "eq", "not-present").run()) == 0

    # Join via a model name:
    assert len(model_a.query().join("Model_B", ("prop1", "eq", "val1")).run()) == 1
    assert len(model_a.query().join("Model_B", ("prop1", "eq", "val2")).run()) == 0

    # Filter via a model instance:
    assert len(model_a.query().filter("prop1", "eq", "val1").run()) == 1
    assert len(model_a.query().filter("prop1", "eq", "not-present").run()) == 0

    # Join via a model instance:
    assert len(model_a.query().join(model_b, ("prop1", "eq", "val1")).run()) == 1
    assert len(model_a.query().join(model_b, ("prop1", "eq", "val2")).run()) == 0

    # use select, join, and filter:
    result = (
        model_a.query()
        .select(model_b)
        .filter("prop1", "eq", "val1")
        .join(model_b, ("prop1", "eq", "val1"))
        .run()
    )

    assert len(result) == 1
    assert len(result[0].items()) == 1
    assert result[0].target.type == model_a.type
    assert result[0].get(model_b).type == model_b.type

    # and with an offset of 1:
    result = (
        model_a.query()
        .select(model_b)
        .filter("prop1", "eq", "val1")
        .join(model_b, ("prop1", "eq", "val1"))
        .offset(1)
        .run()
    )
    assert len(result) == 0

    # and with a limit of 0:
    result = (
        model_a.query()
        .select(model_b)
        .filter("prop1", "eq", "val1")
        .join(model_b, ("prop1", "eq", "val1"))
        .limit(0)
        .run()
    )
    assert len(result) == 0



from .client import Pennsieve
from .config import DEFAULTS as DEFAULT_SETTINGS
from .config import Settings
from .models import (
    BaseNode,
    Collection,
    DataPackage,
    Dataset,
    File,
    LinkedModelProperty,
    Model,
    ModelFilter,
    ModelJoin,
    ModelProperty,
    ModelSelect,
    ModelTemplate,
    Organization,
    Property,
    Record,
    RecordSet,
    Relationship,
    RelationshipSet,
    RelationshipType,
    TimeSeries,
    TimeSeriesAnnotation,
    TimeSeriesChannel,
)

__title__ = "pennsieve"
__version__ = "6.1.2"



import logging
import os


def test_default_log_level():
    from pennsieve import log

    # default log level should be INFO
    base_log = log.get_logger("base_log")
    assert base_log.getEffectiveLevel() == logging.INFO


def test_log_level_based_on_env():
    from pennsieve import log

    # setting env var should change logging level
    os.environ["PENNSIEVE_LOG_LEVEL"] = "WARN"
    warn_log = log.get_logger("warn_log")
    assert warn_log.getEffectiveLevel() == logging.WARN



# -*- coding: utf-8 -*-

# pennsieve-specific
from __future__ import absolute_import, division, print_function
from builtins import object
from future.utils import as_native_str

from . import log
from .api.concepts import (
    ModelRelationshipInstancesAPI,
    ModelRelationshipsAPI,
    ModelsAPI,
    ModelTemplatesAPI,
    RecordsAPI,
)
from .api.core import CoreAPI, OrganizationsAPI, SearchAPI
from .api.data import DataAPI, DatasetsAPI, FilesAPI, PackagesAPI
from .api.timeseries import TimeSeriesAPI
from .api.transfers import IOAPI
from .api.user import UserAPI
from .base import ClientSession
from .config import Settings
from .models import Dataset, ModelTemplate


class Pennsieve(object):
    """
    The client provides an interface to the Pennsieve platform, giving you the
    ability to retrieve, add, and manipulate data.

    Args:
        profile (str, optional): Preferred profile to use
        api_token (str, optional): Preferred api token to use
        api_secret (str, optional): Preferred api secret to use
        jwt (str, optional): A JWT to use for authentication
        headers (dict, optional): A set of global headers to attach to every request
        model_service_host (str, optional): Preferred model service host to use
        host (str, optional): Preferred host to use
        env_override (bool, optional): Should environment variables override settings
        **overrides (dict, optional): Settings to override

    Examples:
        Load the client library and initialize::

            from pennsieve import Pennsieve

            # note: no API token needed if environment variables are set
            ps = Pennsieve()

        Retrieve/modify items on the platform::

            # get some dataset
            ds = ps.get_dataset('my dataset')
            print("Dataset {} has ID={}".format(ds.name, ds.id))

            # list all first-level items in dataset
            for item in ds:
                print("Item: {}".format(item))

            # grab some data package
            pkg = ps.get("N:package:1234-1234-1234-1234")

            # modify a package's name
            pkg.name = "New Package Name"

            # add some property
            pkg.set_property('Room', 'ICU-123')

            # sync changes
            pkg.update()

    Note:
        To initialize your ``Pennsieve`` client without passing any arguments,
        ensure that these environment variables are set:

        - ``PENNSIEVE_API_TOKEN``   (Username of your token pool credentials)
        - ``PENNSIEVE_API_SECRET``  (Password of your token pool credentials)

    """

    def __init__(
        self,
        profile=None,
        api_token=None,
        api_secret=None,
        jwt=None,
        headers=None,
        host=None,
        model_service_host=None,
        env_override=True,
        **overrides
    ):

        self._logger = log.get_logger("pennsieve.client.Pennsieve")

        overrides.update(
            {
                k: v
                for k, v in {
                    "api_token": api_token,
                    "api_secret": api_secret,
                    "api_host": host,
                    "jwt": jwt,
                    "headers": headers,
                    "model_service_host": model_service_host,
                }.items()
                if v is not None
            }
        )
        self.settings = Settings(profile, overrides, env_override)

        if (
            self.settings.api_token is None or self.settings.api_secret is None
        ) and self.settings.jwt is None:
            if self.settings.api_token is None:
                raise Exception(
                    "Error: No API token found. Cannot connect to Pennsieve."
                )
            if self.settings.api_secret is None:
                raise Exception(
                    "Error: No API secret found. Cannot connect to Pennsieve."
                )
            if self.settings.jwt is None:
                raise Exception("Error: No JWT found. Cannot connect to Pennsieve.")

        # direct interface to REST API.
        self._api = ClientSession(self.settings)

        # account
        self._api.authenticate()

        self._api.register(
            CoreAPI,
            OrganizationsAPI,
            DatasetsAPI,
            FilesAPI,
            DataAPI,
            PackagesAPI,
            TimeSeriesAPI,
            SearchAPI,
            IOAPI,
            UserAPI,
            ModelsAPI,
            RecordsAPI,
            ModelRelationshipsAPI,
            ModelRelationshipInstancesAPI,
            ModelTemplatesAPI,
        )

        self._api._context = self._api.organizations.get(self._api._organization)

    @property
    def context(self):
        """
        The current organizational context of the active client.
        """
        return self._api._context

    @property
    def profile(self):
        """
        The profile of the current active user.
        """
        return self._api.profile

    def organizations(self):
        """
        Return all organizations for user.
        """
        return self._api.organizations.get_all()

    def datasets(self):
        """
        Return all datasets for user for an organization (current context).
        """
        self._check_context()
        return self.context.datasets

    def get(self, id, update=True):
        """
        Get any DataPackage or Collection object by ID.

        Args:
            id (str): The ID of the Pennsieve object.

        Returns:
            Object of type/subtype ``DataPackage`` or ``Collection``.
        """
        try:
            return self._api.core.get(id, update=update)
        except:
            self._logger.info(
                "Unable to retrieve object"
                "\n\nAcceptable objects for get() are:"
                "\n - DataPackages"
                "\n - Collections"
                "\n\nUse get_dataset() if trying to retrieve a dataset"
            )

    def create(self, thing):
        """
        Create a object on the platform.
        """
        return self._api.core.create(thing)

    def create_dataset(
        self, name, description=None, automatically_process_packages=False
    ):
        """
        Create a dataset under the active organization.

        Args:
            name (str): The name of the to-be-created dataset

        Returns:
            The created ``Dataset`` object

        """
        self._check_context()
        return self._api.datasets.create(
            Dataset(
                name,
                description=description,
                automatically_process_packages=automatically_process_packages,
            )
        )

    def get_dataset(self, name_or_id):
        """
        Get Dataset by name or ID.

        Args:
            name_or_id (str): the name or the ID of the dataset

        Note:
            When using name, this method gnores case, spaces, hyphens,
            and underscores such that these are equivelent:

              - "My Dataset"
              - "My-dataset"
              - "mydataset"
              - "my_DataSet"
              - "mYdata SET"

        """
        try:
            return self._api.datasets.get(name_or_id)
        except:
            pass

        result = self._api.datasets.get_by_name_or_id(name_or_id)
        if result is None:
            raise Exception("No dataset matching name or ID '{}'.".format(name_or_id))
        return result

    def update(self, thing):
        """
        Update an item on the platform.

        Args:
            thing (object or str): the ID or object to update

        Example::

            my_eeg = ps.get('N:package:1234-1234-1234-1234')
            my_eeg.name = "New EEG Name"
            ps.update(my_eeg)

        Note:
            You can also call update directly on objects, for example::

                my_eeg.name = "New EEG Name"
                my_eeg.update()

        """
        return self._api.core.update(thing)

    def delete(self, *things):
        """
        Deletes items from the platform.

        Args:
            things (list): ID or object of items to delete

        Example::

            # delete a package by passing in object
            ps.delete(my_package)

            # delete a package by ID
            ps.delete('N:package:1234-1234-1234-1235')

            # delete many things (mix of IDs and objects)
            ps.delete(my_package, 'N:collection:1234-1234-1234-1235', another_collection)

        Note:
            When deleting ``Collection``, all child items will also be deleted.

        """
        return self._api.core.delete(*things)

    def move(self, destination, *things):
        """
        Moves item(s) from their current location to the destination.

        Args:
            destination (object or str): The ID of the destination. This must be of type
                type ``Collection`` or ``None``. If destination is ``None``, ``things``
                will be moved to the top of their containing ``Dataset``
            things (list): the IDs or objects to move.

        """
        r = self._api.data.move(destination, *things)

    def search(self, query, max_results=10):
        """
        Find an object on the platform.

        Args:
            query (str): query string to perform search.
            max_results (int, optional): the number of results to return

        Example::

            # find some items belonging to patient123
            for result in  ps.search('patient123'):
                print("found:", result)

        """
        return self._api.search.query(query, max_results=max_results)

    def _check_context(self):
        if self.context is None:
            raise Exception("Must set context before executing method.")

    def get_model_template(self, template_id):
        """
        Get a model template belonging to the current organization.

        Args:
            template_id (str): The ID of the model template object.

        Returns:
            Object of type ModelTemplate.
        """
        return self._api.templates.get(template_id)

    def get_model_templates(self):
        """
        Get all of the model templates belonging to the current organization.

        Returns:
            A list of objects of type ModelTemplate.
        """
        return self._api.templates.get_all()

    def create_model_template(
        self,
        name=None,
        description=None,
        category=None,
        required=None,
        properties=None,
        template=None,
    ):
        """
        Create a new model template belonging to the current organization.

        Args:
            template (ModelTemplate): A ModelTemplate object.
            OR
            name (str): The name of the template
            description (str, optional): A description of the template
            category (str): The category of the template
            required ([str], optional): The names of the required template properties
            properties (dict OR tuples): The template properties

        Example::

            my_template = ModelTemplate(
                schema="http://schema.pennsieve.io/model/draft-01/schema",
                name="Patient",
                description="This is a patient.",
                category="Person",
                properties=dict(
                    name=dict(
                        type="string",
                        description="Name"
                    ),
                    DOB=dict(
                        type="string",
                        format="date",
                        description="Date of Birth"
                    )
                ),
                required=["name"]
            ps.create_model_template(my_template)

        Example::

            ps.create_model_template(
                name="Patient",
                description="This is a patient.",
                category="Person",
                required=["name"],
                properties=[("name", "string"), ("DOB", "date")]
            )

        Returns:
            The created ModelTemplate object.
        """
        if template is None:
            template_to_create = ModelTemplate(
                name=name,
                description=description,
                category=category,
                required=required,
                properties=properties,
            )
        else:
            template_to_create = template
            assert isinstance(
                template_to_create, ModelTemplate
            ), "template must be type ModelTemplate"

        return self._api.templates.create(template=template_to_create)

    def validate_model_template(
        self,
        name=None,
        description=None,
        category=None,
        required=None,
        properties=None,
        template=None,
    ):
        """
        Validate a model template schema.

        Args:
            template (ModelTemplate): A ModelTemplate object.
            OR
            name (str): The name of the template
            description (str, optional): A description of the template
            category (str): The category of the template
            required ([str], optional): The names of the required template properties
            properties (dict OR tuples): The template properties

        Example::

            my_template = ModelTemplate(
                schema="http://schema.pennsieve.io/model/draft-01/schema",
                name="Patient",
                description="This is a patient.",
                category="Person",
                properties=dict(
                    name=dict(
                        type="string",
                        description="Name"
                    ),
                    DOB=dict(
                        type="string",
                        format="date",
                        description="Date of Birth"
                    )
                ),
                required=["name"]
            ps.validate_model_template(my_template)

        Example::

            ps.validate_model_template(
                name="Patient",
                category="Person",
                properties=[("name", "string"), ("Weight", "number")]
            )

        Returns:
            "True" or a list of validation errors
        """
        if template is None:
            template_to_validate = ModelTemplate(
                name=name,
                description=description,
                category=category,
                required=required,
                properties=properties,
            )
        else:
            template_to_validate = template
            assert isinstance(
                template_to_validate, ModelTemplate
            ), "template must be type ModelTemplate"

        return self._api.templates.validate(template=template_to_validate)

    def delete_model_template(self, template_id):
        """
        Deletes a model template from the platform.

        Args:
            template_id (str): The ID of the model template object.

        """
        return self._api.templates.delete(template_id)

    @as_native_str()
    def __repr__(self):
        return "<Pennsieve user='{}' organization='{}'>".format(
            self.profile.email, self.context.name
        )



import datetime
import pdb

import pytest

from pennsieve import TimeSeries, TimeSeriesChannel
from pennsieve.models import TimeSeriesAnnotation, TimeSeriesAnnotationLayer


@pytest.fixture()
def timeseries(client, dataset):
    # create
    ts = TimeSeries("Human EEG")
    assert not ts.exists
    dataset.add(ts)
    assert ts.exists
    assert ts in dataset
    assert ts.type == "TimeSeries"
    assert ts.name.startswith(
        "Human EEG"
    )  # starts with bc duplicate names are appended
    ts2 = client.get(ts.id)
    assert ts2.id == ts.id
    assert ts2.name == ts.name
    assert ts2.type == "TimeSeries"
    del ts2

    # provide to other tests
    yield ts

    # remove
    dataset.remove(ts)
    assert not ts.exists
    assert ts not in dataset


@pytest.fixture()
def timeseries2(client, dataset):
    ts = TimeSeries("Animal EEG")
    dataset.add(ts)
    assert ts.exists
    yield ts
    dataset.remove(ts)


def test_update_timeseries_name(client, timeseries):
    # update timeseries: change name
    timeseries.name = "Monkey EEG"
    timeseries.update()
    timeseries2 = client.get(timeseries.id)
    assert timeseries2.id == timeseries.id
    assert timeseries2.name == timeseries.name
    assert timeseries2.type == "TimeSeries"


def test_timeseries_channels(client, timeseries):
    num_channels = 10

    assert timeseries.exists

    ts2 = client.get(timeseries.id)

    chs = []
    for i in range(num_channels):
        # init
        chname = "Channel-{}".format(i)
        ch = TimeSeriesChannel(name=chname, rate=256, unit="uV")
        assert not ch.exists

        # create
        timeseries.add_channels(ch)
        assert ch.exists
        assert ch.name == chname
        chs.append(ch)

        # use separate request to get channel
        ch.insert_property("key", "value")
        ch2 = [x for x in ts2.channels if x.id == ch.id][0]
        assert ch2.get_property("key") is not None
        assert ch2.name == ch.name
        assert ch2.type == ch.type
        assert ch2.rate == ch.rate
        assert ch2.unit == ch.unit
        del ch2

        # update channel
        ch.name = "{}-updated".format(ch.name)
        ch.rate = 200
        ch.update()

        # use separate request to confirm name change
        ch2 = [x for x in ts2.channels if x.id == ch.id][0]
        assert ch2.name == ch.name
        assert ch2.rate == ch.rate

    # ensure correct number of channels
    channels = timeseries.channels
    assert len(channels) == num_channels

    # get channels (via API)
    ch_ids = [x.id for x in channels]

    # check
    for ch in chs:
        assert ch.id in ch_ids

    ch = channels[0]
    timeseries.remove_channels(ch)
    channels = timeseries.channels
    assert len(channels) == num_channels - 1
    assert ch not in channels
    assert ch.id not in [x.id for x in channels]

    # remove by id
    ch_id = channels[0].id
    timeseries.remove_channels(ch_id)
    channels = timeseries.channels
    assert len(channels) == num_channels - 2
    assert ch_id not in [x.id for x in channels]


# def test_timeseries_annotations(client, timeseries):
#     assert timeseries.exists
#     print("layers = ", timeseries.layers)
#
#     # Create Layer
#     layer1 = TimeSeriesAnnotationLayer(
#         name="test_layer", time_series_id=timeseries.id, description="test_description"
#     )
#     a = layer1.as_dict()
#     assert a["name"] == "test_layer"
#     assert a["description"] == "test_description"
#
#     # Add Layer
#     timeseries.add_layer(layer1)
#     assert layer1.exists
#
#     # Get Layer
#     layer1b = timeseries.get_layer("test_layer")
#     assert layer1b.exists
#     assert layer1b.name == "test_layer"
#     assert layer1.id == layer1b.id
#     assert layer1._api.timeseries is not None
#
#     # Add another layer
#     layer2 = timeseries.add_layer("test_layer2", "test_description2")
#     assert layer2.exists
#
#     layer2copy = timeseries.add_layer("test_layer2")
#     assert layer2copy.id == layer2.id
#
#     # Get Layer
#     layer2b = timeseries.get_layer("test_layer2")
#     assert layer2b.exists
#     assert layer2b.name == "test_layer2"
#     assert layer2.id == layer2b.id
#     assert layer2._api.timeseries is not None
#
#     layers = timeseries.layers
#     assert len(layers) == 2
#     assert set([layers[0].name, layers[1].name]) == set(["test_layer", "test_layer2"])
#
#     # Create channels
#     ch = TimeSeriesChannel(
#         name="test_channel", rate=256, unit="uV", start=1, end=60 * 1e6
#     )
#
#     # Create annotation over one channel
#     # create
#     timeseries.add_channels(ch)
#     assert ch.exists
#     assert ch.id in [x.id for x in timeseries.channels]
#     annot = TimeSeriesAnnotation(
#         label="test_label",
#         channel_ids=timeseries.channels[0].id,
#         start=timeseries.channels[0].start,
#         end=timeseries.channels[0].start + 1 * 1e6,
#     )
#     # Add Annotation
#     layer1.add_annotations(annot)
#     assert annot.exists
#
#     # get annotations
#     annotb = layer1.annotations()
#     assert annotb[0].label == annot.label
#
#     annotc = client._api.timeseries.get_annotation(timeseries, layer1, annot)
#     assert annotc.label == annot.label
#
#     # Create annotation over multiple channels
#     ch2 = TimeSeriesChannel(
#         name="test_channel", rate=256, unit="uV", start=1, end=60 * 1e6
#     )
#
#     timeseries.add_channels(ch2)
#     channels = timeseries.channels
#     ch_ids = [x.id for x in channels]
#     assert ch2.exists
#     assert ch2.id in ch_ids
#     assert ch.id in ch_ids
#     for ch in channels:
#         assert ch.rate == 256
#         assert ch.exists
#
#     # add annotation over two channels
#     channel_ids = [timeseries.channels[x].id for x in range(len(timeseries.channels))]
#
#     annot2 = layer1.insert_annotation(
#         annotation="test_label2",
#         channel_ids=channel_ids,
#         start=timeseries.channels[0].start + 1 * 1e6,
#         end=timeseries.channels[0].start + 2 * 1e6,
#     )
#
#     assert annot2.exists
#
#     annot_gen = layer1.iter_annotations(1)
#     annot = next(annot_gen)
#     assert annot[0].label == "test_label"
#     next_annot = next(annot_gen)
#     assert next_annot[0].label == "test_label2"
#
#     ### TEST ANNOTATION COUNTS
#     layer1_expected_counts = [
#         {"start": 0, "end": 250000, "value": 1.0},
#         {"start": 250000, "end": 500000, "value": 1.0},
#         {"start": 500000, "end": 750000, "value": 1.0},
#         {"start": 750000, "end": 1000000, "value": 1.0},
#         {"start": 1000000, "end": 1250000, "value": 2.0},
#         {"start": 1250000, "end": 1500000, "value": 1.0},
#         {"start": 1500000, "end": 1750000, "value": 1.0},
#         {"start": 1750000, "end": 2000000, "value": 1.0},
#         {"start": 2000000, "end": 2000001, "value": 1.0},
#     ]
#
#     def _sort_counts(counts):
#         return sorted(counts, key=lambda c: c["start"])
#
#     assert (
#         _sort_counts(
#             timeseries.annotation_counts(
#                 start=timeseries.channels[0].start * 1e6,
#                 end=timeseries.channels[0].start + 2 * 1e6,
#                 layers=[layer1],
#                 period="0.25s",
#             )[str(layer1.id)]
#         )
#         == layer1_expected_counts
#     )
#     assert (
#         _sort_counts(
#             layer1.annotation_counts(
#                 start=timeseries.channels[0].start * 1e6,
#                 end=timeseries.channels[0].start + 2 * 1e6,
#                 period="0.25s",
#             )[str(layer1.id)]
#         )
#         == layer1_expected_counts
#     )
#
#     ### TEST DELETION
#
#     annot3 = TimeSeriesAnnotation(
#         label="test_label3",
#         channel_ids=channel_ids,
#         start=timeseries.channels[0].start + 1 * 1e6,
#         end=timeseries.channels[0].start + 2 * 1e6,
#     )
#     layer1.add_annotations([annot2, annot3])
#     annot3 = timeseries.add_annotations(layer=layer1, annotations=annot3)
#     assert annot3.exists
#     annot3.delete()
#     assert not annot3.exists
#
#     annot4 = timeseries.insert_annotation(
#         layer=layer1,
#         annotation="test_label3",
#         start=timeseries.channels[0].start + 1 * 1e6,
#         end=timeseries.channels[0].start + 2 * 1e6,
#     )
#     assert annot4.exists
#     annot4.delete()
#     assert not annot4.exists
#
#     annot5 = timeseries.insert_annotation(
#         layer="test_layer4",
#         annotation="test_label3",
#         start=timeseries.channels[0].start + 1 * 1e6,
#         end=timeseries.channels[0].start + 2 * 1e6,
#     )
#     assert annot5.exists
#     annot5.delete()
#     assert not annot5.exists
#
#     layer1.add_annotations([annot2, annot3])
#     assert annot2.exists
#     assert annot3.exists
#
#     # test datetime input
#     annot4 = timeseries.insert_annotation(
#         layer="test_layer4",
#         annotation="test_label3",
#         start=datetime.datetime.utcfromtimestamp(
#             (timeseries.channels[0].start + 1 * 1e6) / 1e6
#         ),
#         end=datetime.datetime.utcfromtimestamp(timeseries.channels[0].start + 2 * 1e6),
#     )
#     assert annot4.exists
#     annot4.delete()
#     assert not annot4.exists
#
#     layer = timeseries.get_layer("test_layer4")
#     assert layer.exists
#     layer.delete()
#     assert not layer.exists
#
#     # delete annotations
#     annot[0].delete()
#     assert not annot[0].exists
#
#     assert timeseries.exists
#     timeseries.delete_layer(layer1)
#     assert not layer1.exists
#
#     assert layer2.exists
#     layer2.delete()
#     assert not layer2.exists
#
#
# def test_timeseries_annotations_can_get_more_than_default_limit(timeseries):
#     ch = TimeSeriesChannel(name="test_channel", rate=256, unit="uV", start=0, end=2e6)
#     timeseries.add_channels(ch)
#
#     layer = TimeSeriesAnnotationLayer(
#         name="limit_layer",
#         time_series_id=timeseries.id,
#         description="layer with many small annotations",
#     )
#     timeseries.add_layer(layer)
#
#     for i in range(200):
#         annotation = TimeSeriesAnnotation(
#             label="annotation_{}".format(i),
#             channel_ids=ch.id,
#             start=i * 1e4,
#             end=(i + 1) * 1e4,
#         )
#         layer.add_annotations(annotation)
#
#     # Should retrieve all annotations
#     assert len(layer.annotations()) == 200
#
#     # All of these annotations should be in the first 10-second window
#     annots = next(layer.iter_annotations())
#     assert len(annots) == 200
#
#     layer.delete()
#
#
# def test_timeseries_annotations_check_that_channels_belong_to_series(
#     timeseries, timeseries2
# ):
#     ch = TimeSeriesChannel(name="test_channel", rate=256, unit="uV", start=0, end=1e6)
#     timeseries.add_channels(ch)
#     print(timeseries, timeseries2)
#
#     ch2 = TimeSeriesChannel(
#         name="test_channel_2", rate=256, unit="uV", start=0, end=1e6
#     )
#     timeseries2.add_channels(ch2)
#
#     layer = TimeSeriesAnnotationLayer(
#         name="a_layer", time_series_id=timeseries.id, description="a layer"
#     )
#     timeseries.add_layer(layer)
#
#     with pytest.raises(Exception):
#         layer.annotations(channels=[ch, ch2])
#
#     with pytest.raises(Exception):
#         next(layer.iter_annotations(channels=[ch, ch2]))
#
#     with pytest.raises(Exception):
#         layer.annotation_counts(0, 1e6, period="0.1s", channels=[ch, ch2])
#
#
# def test_timeseries_segments(client, timeseries):
#     """
#     No valid timeseries data to test against, but we can test the API calls.
#     """
#     channels = [
#         TimeSeriesChannel(name="ch 1", rate=256, unit="uV"),
#         TimeSeriesChannel(name="ch 2", rate=111.123, unit="uV"),
#     ]
#     timeseries.add_channels(*channels)
#
#     for ch in timeseries.channels:
#         ch.end = ch.start + 5e6
#         ch.update()
#
#         ch.segments()
#         ch.segments(start=ch.start + 1e6)
#         ch.segments(stop=ch.end - 1e6)
#         ch.segments(stop=ch.end - 1e6, gap_factor=3)
#         ch.segments(stop=ch.end - 1e6, gap_factor=33.33)
#         ch.segments(stop=ch.end - 1e6, gap_factor=0.5)
#
#     timeseries.segments()
#     timeseries.segments(gap_factor=3)
#     timeseries.segments(gap_factor=33.33)
#     timeseries.segments(gap_factor=0.5)
#
#     # TODO: API should soon error when value is non-numeric.
#     #       Uncomment below when API is changed.
#
#     with pytest.raises(Exception):
#         timeseries.segments(gap_factor="should be int")
#
#     with pytest.raises(Exception):
#         ch.segments(gap_factor="should be int")



import datetime
import os

import pytest
import requests
from dateutil.parser import parse

from pennsieve import Pennsieve
from pennsieve.base import UnauthorizedException
from pennsieve.models import (
    BaseNode,
    DataPackage,
    Dataset,
    File,
    PublishInfo,
    TeamCollaborator,
    UserCollaborator,
)

from .utils import get_test_client


def test_basenode(client, dataset):
    node1 = dataset
    node2 = client.get_dataset(dataset.id)
    node3 = BaseNode()
    assert node1 == node2
    assert node1 != node3
    assert node1 != object()
    assert [node1, node2] == [node2, node1]


def test_update_dataset(client, dataset, session_id):
    # update name of dataset
    ds_name = "Same Dataset, Different Name {}".format(session_id)
    dataset.name = ds_name
    dataset.update()
    ds2 = client.get_dataset(dataset.id)
    assert ds2.id == dataset.id
    assert isinstance(ds2.int_id, int)
    assert ds2.int_id == dataset.int_id
    assert ds2.name == ds_name
    assert ds2.owner_id == client.profile.id


def test_dataset_status_log(client, dataset):
    dataset_status_log = dataset.status_log()
    assert dataset_status_log.limit == 25
    assert dataset_status_log.offset == 0
    assert dataset_status_log.total_count == 1
    assert len(dataset_status_log.entries) == 1
    assert dataset_status_log.entries[0].user.node_id == client.profile.id
    assert dataset_status_log.entries[0].user.first_name == client.profile.first_name
    assert dataset_status_log.entries[0].user.last_name == client.profile.last_name
    assert dataset_status_log.entries[0].updated_at == parse(dataset.created_at)
    assert dataset_status_log.entries[0].status.id is not None
    assert dataset_status_log.entries[0].status.name is not None
    assert dataset_status_log.entries[0].status.display_name is not None


def test_status_is_readonly(client, dataset):
    with pytest.raises(AttributeError) as excinfo:
        dataset.status = "New Thing"
        dataset.update()
    assert "Dataset.status is read-only." in str(excinfo.value)


def test_tags_is_list_of_strings_only(client, dataset):
    with pytest.raises(AttributeError) as excinfo:
        dataset.tags = "New Thing"
    assert "Dataset.tags should be a list of strings." in str(excinfo.value)
    with pytest.raises(AttributeError) as excinfo:
        dataset.tags = [1, 2, 3]
    assert "Dataset.tags should be a list of strings." in str(excinfo.value)
    dataset.tags = ["a", "b", "c"]
    dataset.update()
    dataset_from_platform = client.get_dataset(dataset.id)
    assert dataset_from_platform.tags == ["a", "b", "c"]


def test_datasets(client, dataset):
    ds_items = len(dataset)

    # create package locally
    pkg = DataPackage("Child of Dataset", package_type="Text")

    assert not pkg.exists
    assert pkg not in dataset

    # add package to dataset
    dataset.add(pkg)

    assert pkg.exists
    assert pkg in dataset
    assert pkg.id in dataset
    assert len(dataset) == ds_items + 1

    # remove from dataset
    dataset.remove(pkg)

    assert not pkg.exists
    assert pkg not in dataset
    assert len(dataset) == ds_items

    # can't create dataset with same name
    with pytest.raises(Exception):
        client.create_dataset(dataset.name)


def test_packages_create_delete(client, dataset):

    # init
    pkg = DataPackage("Some MRI", package_type="MRI")
    assert not pkg.exists

    # create
    dataset.add(pkg)
    assert pkg.exists
    assert pkg.id is not None
    assert pkg.name == "Some MRI"
    assert pkg.owner_id == client.profile.int_id

    # TODO: (once we auto-include in parent)
    assert pkg in dataset

    # update package name
    pkg.name = "Some Other MRI"
    pkg = client.update(pkg)

    pkg2 = client.get(pkg.id)

    assert pkg2.name == "Some Other MRI"
    assert pkg2.id == pkg.id
    assert pkg2.owner_id == client.profile.int_id

    # delete all packages
    client.delete(pkg)

    assert not pkg.exists

    pkg = DataPackage("Something else", package_type="TimeSeries")
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists
    pid = pkg.id
    pkg.delete()
    assert not pkg.exists

    pkg2 = client.get(pid)
    assert pkg2 is None

    # TODO: (once we auto-remove from parent)
    # assert pkg not in dataset


def test_package_type_count(client, dataset):
    n = dataset.package_count()
    pkg = DataPackage("Some MRI", package_type="MRI")
    assert not pkg.exists
    # create
    dataset.add(pkg)
    assert pkg.exists
    client.update(pkg)

    pkg = DataPackage("Something else", package_type="TimeSeries")
    assert not pkg.exists
    dataset.add(pkg)
    assert pkg.exists
    client.update(pkg)

    m = dataset.package_count()
    assert m == n + 2


def test_publish_info(client, dataset):
    publish_info = dataset.published()
    assert publish_info.status == "NOT_PUBLISHED"
    assert publish_info.version_count == 0
    assert publish_info.last_published == None
    assert publish_info.doi == None


def test_owner(client, dataset):
    owner = dataset.owner()
    assert owner.email == client.profile.email


def test_collaborator_user(client, dataset):
    collaborators = dataset.user_collaborators()
    assert len(collaborators) == 1
    assert collaborators[0].email == client.profile.email


def test_collaborator_team(client, dataset):
    collaborators = dataset.team_collaborators()
    assert len(collaborators) == 0


def test_properties(client, dataset):

    pkg = DataPackage("Some Video", package_type="Video")
    assert not pkg.exists

    dataset.add(pkg)
    assert pkg.exists

    pkg.insert_property("my-key", "my-value")
    pkg2 = client.get(pkg)
    print("properties =", pkg2.properties)
    assert pkg2.id == pkg.id
    assert pkg2.get_property("my-key").data_type == "string"
    assert pkg2.get_property("my-key").value == "my-value"

    explicit_ptypes = {
        "my-int1": ("integer", 123123),
        "my-int2": ("integer", "123123"),
        "my-float": ("double", 123.123),
        "my-float2": ("double", "123.123"),
        "my-float3": ("double", "123123"),
        "my-date": ("date", 1488847449697),
        "my-date2": ("date", 1488847449697.123),
        "my-date3": ("date", datetime.datetime.now()),
        "my-string": ("string", "my-123123"),
        "my-string2": ("string", "123123"),
        "my-string3": ("string", "123123.123"),
        "my-string4": ("string", "According to plants, humans are blurry."),
    }
    for key, (ptype, val) in explicit_ptypes.items():
        pkg.insert_property(key, val, data_type=ptype)
        assert pkg.get_property(key).data_type == ptype

    inferred_ptypes = {
        "my-int1": ("integer", 123123),
        "my-int2": ("integer", "123123"),
        "my-float1": ("double", 123.123),
        "my-float2": ("double", "123.123"),
        "my-date": ("date", datetime.datetime.now()),
        "my-string": ("string", "i123123"),
        "my-string2": ("string", "#1231"),
    }
    for key, (ptype, val) in inferred_ptypes.items():
        pkg.insert_property(key, val)
        prop = pkg.get_property(key)
        assert prop.data_type == ptype

    # remove property
    pkg.remove_property("my-key")
    assert pkg.get_property("my-key") is None

    pkg2 = client.get(pkg.id)
    assert pkg2.get_property("my-key") is None


def test_can_remove_multiple_items(dataset):
    pkg1 = DataPackage("Some MRI", package_type="MRI")
    dataset.add(pkg1)
    pkg1.update()
    pkg2 = DataPackage("Some Video", package_type="Video")
    dataset.add(pkg2)
    pkg2.update()
    assert pkg1 in dataset.items
    assert pkg2 in dataset.items

    dataset.remove(pkg1)
    dataset.remove(pkg2)
    assert pkg1 not in dataset.items
    assert pkg2 not in dataset.items


def test_timeout():
    with pytest.raises(requests.exceptions.Timeout):
        # initial authentication calls should time out
        get_test_client(max_request_time=0.00001)


def test_client_host_overrides():
    host = "http://localhost"
    # fails authentication in Pennsieve.__init__
    with pytest.raises(requests.exceptions.RequestException):
        ps = Pennsieve(host=host)

    ps = Pennsieve(model_service_host=host)
    assert ps.settings.model_service_host == host


def test_exception_raise():
    ps = Pennsieve()
    with pytest.raises(Exception) as excinfo:
        ps._api._call("get", "/datasets/plop")
    assert "plop not found" in str(excinfo.value)


def test_client_global_headers():
    global_headers = {
        "X-Custom-Header1": "Custom Value",
        "X-Custom-Header2": "Custom Value2",
    }
    ps = Pennsieve(headers=global_headers)
    assert ps.settings.headers == global_headers

    for header, header_value in global_headers.items():
        assert header in ps._api._session.headers
        assert ps._api._session.headers[header] == header_value



from __future__ import absolute_import, division, print_function

import io
import re
from os import path

from setuptools import find_packages, setup

with io.open("pennsieve/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

here = path.abspath(path.dirname(__file__))

with io.open(path.join(here, "requirements.txt"), mode="r", encoding="utf-8") as f:
    reqs = [line.strip() for line in f if not line.startswith("#")]

# Get the long description from the README file
with io.open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pennsieve",
    version=version,
    author="Pennsieve, Inc.",
    author_email="joost@pennsieve.com",
    description="Python client for the Pennsieve Platform",
    long_description=long_description,
    packages=find_packages(),
    package_dir={"pennsieve": "pennsieve"},
    setup_requires=["cython"],
    install_requires=reqs,
    extras_require={
        "data": ["numpy>=1.13", "pandas>=0.20"],
    },
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4.0",
    entry_points={
        "console_scripts": ["pennsieve-profile=pennsieve.cli.pennsieve_profile:main"]
    },
    license="",
    keywords="pennsieve client rest api",
    url="https://github.com/Pennsieve/pennsieve-python",
    project_urls={
        "Pennsieve": "https://www.pennsieve.com",
        # Do not remove or rename this "Documentation" URL: it is used by
        # the Sphinx theme to generate links to previous doc versions.
        # However, you can change the URL with no issues.
        "Documentation": "https://developer.pennsieve.io/python",
        "Bug Reports": "https://github.com/Pennsieve/pennsieve-python/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from future.utils import string_types

import datetime
import math

import requests

from pennsieve import log
from pennsieve.api.base import APIBase
from pennsieve.extensions import numpy as np
from pennsieve.extensions import pandas as pd
from pennsieve.extensions import require_extension
from pennsieve.models import (
    BaseDataNode,
    Collection,
    DataPackage,
    Dataset,
    File,
    Organization,
    PublishInfo,
    StatusLogEntry,
    StatusLogResponse,
    TeamCollaborator,
    User,
    UserCollaborator,
)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Dataset
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class DatasetsAPI(APIBase):
    """
    Interface for managing datasets on the platform.
    """

    base_uri = "/datasets"
    name = "datasets"

    def get(self, ds):
        id = self._get_id(ds)
        resp = self._get(self._uri("/{id}", id=id))
        return Dataset.from_dict(resp, api=self.session)

    def published(self, ds):
        id = self._get_id(ds)
        resp = self._get(self._uri("/{id}/published", id=id))
        if resp["status"] == "PUBLISH_SUCCEEDED":
            resp["latest_doi"] = self._get(self._uri("/{id}/doi", id=id))["doi"]
        return PublishInfo.from_dict(resp)

    def package_count(self, ds):
        id = self._get_id(ds)
        resp = self._get(self._uri("/{id}/packageTypeCounts", id=id))
        file_count = 0
        for key, value in resp.items():
            file_count += value
        return file_count

    def status_log(self, ds, limit, offset):
        id = self._get_id(ds)
        resp = self._get(
            self._uri(
                "/{id}/status-log?limit={limit}&offset={offset}",
                id=id,
                limit=limit,
                offset=offset,
            )
        )
        return StatusLogResponse.from_dict(resp)

    def team_collaborators(self, ds):
        id = self._get_id(ds)
        resp = self._get(self._uri("/{id}/collaborators/teams", id=id))
        return [TeamCollaborator.from_dict(t) for t in resp]

    def user_collaborators(self, ds):
        id = self._get_id(ds)
        resp = self._get(self._uri("/{id}/collaborators/users", id=id))
        return [UserCollaborator.from_dict(u) for u in resp]

    def get_packages_by_filename(self, ds, filename):
        id = self._get_id(ds)
        resp = self._get(
            self._uri("/{id}/packages?filename={filename}", id=id, filename=filename)
        )
        return [
            DataPackage.from_dict(p, api=self.session) for p in resp.get("packages")
        ]

    def owner(self, ds):
        return next(
            iter(filter(lambda x: x.role == "owner", self.user_collaborators(ds)))
        )

    def get_by_name_or_id(self, name_or_id):
        """
        Get Dataset by name or ID.

        When using name, this ignores case, spaces, hyphens, and underscores
        such that these are equivalent:

          - "My Dataset"
          - "My-dataset"
          - "mydataset"
          - "my_DataSet"
          - "mYdata SET"

        """

        def name_key(n):
            return n.lower().strip().replace(" ", "").replace("_", "").replace("-", "")

        search_key = name_key(name_or_id)

        def is_match(ds):
            return (name_key(ds.name) == search_key) or (ds.id == name_or_id)

        matches = [ds for ds in self.get_all() if is_match(ds)]
        return matches[0] if matches else None

    def get_all(self):
        resp = self._get(self._uri("/"))
        return [Dataset.from_dict(ds, api=self.session) for ds in resp]

    def create(self, ds):
        """
        Create a dataset on the platform
        """
        if self.get_by_name_or_id(ds.name) is not None:
            raise Exception("Dataset with name {} already exists".format(ds.name))

        resp = self._post("", json=ds.as_dict())
        return Dataset.from_dict(resp, api=self.session)

    def update(self, ds):
        """
        Update a dataset on the platform
        """
        id = self._get_id(ds)
        resp = self._put(self._uri("/{id}", id=id), json=ds.as_dict())
        return Dataset.from_dict(resp, api=self.session)

    def delete(self, ds):
        """
        Delete a dataset on the platform
        """
        id = self._get_id(ds)
        resp = self._del(self._uri("/{id}", id=id))
        ds.id = None
        return resp


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Data
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class DataAPI(APIBase):
    """
    Interface for lower-level data operations on the platform.
    """

    base_uri = "/data"
    name = "data"

    def update_properties(self, thing):
        """
        Update properties for an object/package on the platform.
        """
        path = self._uri("/{id}/properties", id=thing.id)
        body = {"properties": [m.as_dict() for m in thing.properties]}

        return self._put(path, json=body)

    def delete(self, *things):
        """
        Deletes objects from the platform
        """
        ids = list(set([self._get_id(x) for x in things]))
        r = self._post("/delete", json=dict(things=ids))
        if len(r["success"]) != len(ids):
            failures = [f["id"] for f in r["failures"]]
            print("Unable to delete objects: {}".format(failures))

        for thing in things:
            if isinstance(thing, BaseDataNode):
                thing.id = None

        return r

    def move(self, destination, *things):
        """
        Moves objects to the destination package
        """
        ids = [self._get_id(x) for x in things]
        # if destination is None, things will get moved into their containing dataset
        dest = self._get_id(destination) if destination is not None else None
        return self._post("/move", json=dict(things=ids, destination=dest))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Packages
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class PackagesAPI(APIBase):
    """
    Interface for task/workflow objects on Pennsieve platform
    """

    base_uri = "/packages"
    name = "packages"

    def create(self, pkg):
        """
        Create data package on platform
        """
        resp = self._post("", json=pkg.as_dict())
        pkg = self._get_package_from_data(resp)
        return pkg

    def update(self, pkg, **kwargs):
        """
        Update package on platform
        """
        d = pkg.as_dict()
        d.update(kwargs)
        d.pop("state", None)
        resp = self._put(self._uri("/{id}", id=pkg.id), json=d)
        pkg = self._get_package_from_data(resp)
        return pkg

    def get(self, pkg, include=None):
        """
        Get package object

        pkg:     can be DataPackage ID or DataPackage object.
        include: list of fields to force-include in response (if available)
        """
        pkg_id = self._get_id(pkg)

        params = None
        if include is not None:
            if isinstance(include, string_types):
                params = {"include": include}
            if hasattr(include, "__iter__"):
                params = {"include": ",".join(include)}

        resp = self._get(self._uri("/{id}", id=pkg_id), params=params)

        # TODO: cast to specific DataPackages based on `type`
        pkg = self._get_package_from_data(resp)
        return pkg

    def process(self, pkg):
        """
        Process a package that has been successfully uploaded but not yet processed
        """
        try:
            self._put(self._uri("/{id}/process", id=pkg.id))
            return True
        except requests.exceptions.HTTPError as error:
            response = error.response
            status_code = response.status_code
            message = response.json().get("message")

            if status_code == requests.codes.bad_request and message is not None:
                raise Exception(message)
            else:
                raise error

    def get_sources(self, pkg):
        """
        Returns the sources of a DataPackage. Sources are the raw, unmodified
        files (if they exist) that contains the package's data.
        """
        pkg_id = self._get_id(pkg)
        resp = self._get(self._uri("/{id}/sources", id=pkg_id))
        for r in resp:
            r["content"].update(dict(pkg_id=pkg_id))

        return [File.from_dict(r, api=self.session) for r in resp]

    def get_files(self, pkg):
        """
        Returns the files of a DataPackage. Files are the possibly modified
        source files (e.g. converted to a different format), but they could also
        be the source files themselves.
        """
        pkg_id = self._get_id(pkg)
        resp = self._get(self._uri("/{id}/files", id=pkg_id))
        for r in resp:
            r["content"].update(dict(pkg_id=pkg_id))

        return [File.from_dict(r, api=self.session) for r in resp]

    def get_view(self, pkg):
        """
        Returns the object(s) used to view the package. This is typically a set of
        file objects, that may be the DataPackage's sources or files, but could also be
        a unique object specific for the viewer.
        """
        pkg_id = self._get_id(pkg)
        resp = self._get(self._uri("/{id}/view", id=pkg_id))
        for r in resp:
            r["content"].update(dict(pkg_id=pkg_id))

        return [File.from_dict(r, api=self.session) for r in resp]

    def get_presigned_url_for_file(self, pkg, file):
        args = dict(pkg_id=self._get_id(pkg), file_id=self._get_id(file))
        resp = self._get(self._uri("/{pkg_id}/files/{file_id}", **args))
        if "url" in resp:
            return resp["url"]
        else:
            raise Exception("Unable to get URL for file ID = {}".format(file_id))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Files
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class FilesAPI(APIBase):
    """
    Interface for managing file object in Pennsieve
    """

    base_uri = "/files"
    name = "files"

    def create(self, file, destination=None):
        """
        Creates a file under the given destination or its current parent
        """
        container = file.parent if destination is None else destination

        body = file.as_dict()
        body["container"] = container

        response = self._post("", json=body)

        return File.from_dict(response, api=self.session)

    def update(self, file):
        """
        Update a file on the platform
        """
        loc = self._uri("/{id}", id=file.id)
        return self._put(loc, json=file.as_dict())

    def url(self, file):
        """
        Get pre-signed URL for File object
        """
        loc = self._uri("/{id}", id=file.id)
        return self._get(loc)["url"]



from uuid import uuid4

import pytest

from pennsieve.models import LinkedModelProperty, ModelProperty
from tests.utils import create_test_dataset


def make_id():
    return str(uuid4()).replace("-", "_")


### Testing linked properties locally:
def test_make_linked_property(dataset):
    # make a new model
    model = dataset.create_model(
        "my_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )

    # create a linked property linking to that model,
    # and make sure the link initialized correctly
    link1 = LinkedModelProperty("link1", model, "1st linked property")

    dct1 = link1.as_dict()
    assert dct1["name"] == "link1"
    assert dct1["displayName"] == "1st linked property"
    assert dct1["to"] == model.id

    # make a linked property from a dict,
    # and make sure it initialized correctly
    link2 = LinkedModelProperty.from_dict(
        {
            "link": {
                "name": "link2",
                "displayName": "2nd linked property",
                "to": model.id,
                "id": "XXX-XXX-XXX",
                "position": 0,
            }
        }
    )
    assert link2.id == "XXX-XXX-XXX"
    assert link2.position == 0

    dct2 = link2.as_dict()
    assert dct2["name"] == "link2"
    assert dct2["displayName"] == "2nd linked property"
    assert dct2["to"] == model.id


### Testing linked property API methods:
def test_add_linked_property(dataset):
    # Create two models and link one to the other
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    source.add_linked_property("link", target, "my linked property")

    # Make sure newly created link is accessible through the API
    assert any(
        l.name == "link" and l.target == target.id
        for l in dataset.get_topology()["linked_properties"]
    )
    assert "link" in source.linked

    # Prevent user from adding duplicate linked properties
    with pytest.raises(Exception):
        source.add_linked_property("link", source, "duplicate linked property")
    assert not any(
        l.display_name == "duplicate linked property"
        for l in dataset.get_topology()["linked_properties"]
    )


def test_add_linked_property_bulk(dataset):
    # Link one model to three others
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    link1 = LinkedModelProperty("link1", target, "bulk-added")
    link2 = LinkedModelProperty("link2", target, "bulk-added")
    link3 = LinkedModelProperty("link3", target, "bulk-added")
    source.add_linked_properties([link1, link2, link3])

    # Make sure newly created links are accessible through the API
    bulk = [
        x
        for x in dataset.get_topology()["linked_properties"]
        if x.display_name == "bulk-added"
    ]
    assert len(bulk) == 3
    assert len(source.linked) == 3


def test_edit_linked_property(dataset):
    # make a model and add a linked property
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    link = source.add_linked_property("link", target, "my linked property")

    # edit the linked property and update model
    link.display_name = "updated linked property"
    link.position = 99
    source.update()

    # Make sure changes were saved
    new_link = source.get_linked_property("link")
    assert new_link.position == 99
    assert new_link.display_name == "updated linked property"


def test_delete_linked_property(dataset):
    # make a model and add a linked property
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    source.add_linked_property("link", target, "my linked property")

    # delete the link
    source.remove_linked_property("link")

    # Make sure changes were saved
    assert "link" not in source.linked


def test_retrieve_linked_properties(dataset):
    # make a model and add a linked property
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    source.add_linked_property("link", target, "my linked property")

    new_source = dataset.get_model(source.type)
    assert "link" in new_source.linked


### Testing linked property values
def test_add_link(dataset):
    # make a model and add a linked property
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    prop = source.add_linked_property("link", target, "my linked property")

    # make records and link them
    source_rec = source.create_record({"name": "source_record"})
    target_rec = target.create_record({"name": "target_record"})
    link = source_rec.add_linked_value(target_rec, prop)
    assert link.id == source_rec.get_linked_value("link").id

    # prevent duplicate links from being created
    target_rec2 = target.create_record({"name": "second_target"})
    link2 = source_rec.add_linked_value(target_rec2, prop)
    links = source_rec.get_linked_values()
    assert len(links) == 1
    assert links[0].target_record_id == target_rec2.id


def test_get_link(dataset):
    # make a model and add a linked property
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    prop = source.add_linked_property("link", target, "my linked property")

    # make records and link them
    source_rec = source.create_record({"name": "source_record"})
    target_rec = target.create_record({"name": "target_record"})
    link = source_rec.add_linked_value(target_rec, prop)
    link2 = source_rec.get_linked_value("link")
    link3 = source_rec.get_linked_value(link.id)
    assert link.id == link2.id == link3.id
    assert link.source_model.id == source.id
    assert link.target_model.id == target.id


def test_remove_link(dataset):
    # make a model and add a linked property
    source = dataset.create_model(
        "source_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    target = dataset.create_model(
        "target_model_{}".format(make_id()), schema=[ModelProperty("name", title=True)]
    )
    prop = source.add_linked_property("link", target, "my linked property")

    # make records and link them
    source_rec = source.create_record({"name": "source_record"})
    target_rec = target.create_record({"name": "target_record"})
    link = source_rec.add_linked_value(target_rec, prop)

    # delete the link
    source_rec.delete_linked_value(link.id)
    assert not any(
        link.target == target_rec.id for link in source_rec.get_linked_values()
    )
